import sys

def format_cp2k_inp(indent: int = 2):
    """
    cp2k uses `&keyword` to start a block, and `&end` to end a block, for example,
    &block1 xxx
      clause 1
      clause 2
      &block2
       clause 3
      &end block2
    &end block1

    To format the file, this function will:
    1. remove leding and trailing whitespaces
    2. autoindent the file
    3. print the file to stdout
    """

    indent_level = 0
    for line in sys.stdin.readlines():
        assert indent_level >= 0, 'indent level should not be negative'
        line = line.strip()
        line_low = line.lower()
        if line_low.startswith('&end'):
            indent_level -= indent
            print(' ' * indent_level + line)
        else:
            print(' ' * indent_level + line)
            if line_low.startswith('&'):
                indent_level += indent

cmd_entry = {
    'format_cp2k_inp': format_cp2k_inp,
}



from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.script import BashTemplate
from ai2_kit.core.script import BashScript, BashStep
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import dict_nested_get, expand_globs, dump_json, list_split, flatten
from ai2_kit.core.pydantic import BaseModel
from ai2_kit.tool.dpdata import set_fparam, register_data_types

from typing import List, Tuple, Optional
from dataclasses import dataclass
from itertools import groupby
import os
import sys
import copy
import random
import dpdata

from .iface import ICllTrainOutput, BaseCllContext, TRAINING_MODE
from .data import DataFormat, get_data_format
from .dpff import set_dplr_ext_from_cp2k_output

from .constant import (
    DP_CHECKPOINT_FILE,
    DP_DISP_FILE,
    DP_PROFILING_FILE,
    DP_INPUT_FILE,
    DP_FROZEN_MODEL,
    DP_ORIGINAL_MODEL,
)

logger = get_logger(__name__)


class CllDeepmdInputConfig(BaseModel):

    class DwTraningConfig(BaseModel):
        """
        Options for deep wannier model training
        """
        input_template: dict
        """
        Deepmd input template.
        Ref: https://docs.deepmodeling.com/projects/deepmd/en/master/model/dplr.html
        """

    train_dw: Optional[DwTraningConfig] = None
    """
    Options for deep wannier model training, if None, then deep wannier model will not be trained.
    """

    model_num: int = 4
    """
    Total number of models to train.
    """
    init_dataset: List[str] = []
    """
    Dataset used to initialize training.
    """
    input_template: dict = dict()
    """
    Deepmd input template.
    """
    compress_model: bool = False
    """
    Whether to compress model after training.
    """
    pretrained_model: Optional[str] = None
    """
    Pretrained model used to finetune.
    """
    isolate_outliers: bool = False
    """
    If isolate_outliers is enabled, then outlier data will be separated from training data.
    """
    outlier_f_cutoff: float = 10.
    """
    The threshold of force magnitude to determine whether a data is outlier.
    """
    outlier_weight: float = 0.003
    """
    The weight of outlier data in training data.
    """

    fixture_models: List[str] = []
    """
    Fixture models used to initialize training, support glob pattern.
    If this is not empty, then the whole training process will be skipped.
    This feature is useful for debugging, or explore more structures without training.
    The models should be on the remote executor.
    The name fixture is used as the concept of fixture in pytest.
    """

    group_by_formula: bool = False
    """
    Grouping dataset by formula
    If this is enabled, then the dataset will be grouped by formula.
    Otherwise, the dataset will be grouped by ancestor.

    Set this to True when you have multiple structures with the same ancestor.
    """

    init_from_previous: bool = False
    """
    Use the previous models to initialize the current training,
    which can speed up the training process.
    """

    ignore_error: bool = False


class CllDeepmdContextConfig(BaseModel):
    script_template: BashTemplate
    dp_cmd: str = 'dp'
    concurrency: int = 5


@dataclass
class CllDeepmdInput:
    config: CllDeepmdInputConfig
    mode: TRAINING_MODE
    type_map: List[str]
    sel_type: Optional[List[int]]
    old_dataset: List[Artifact]  # training data used by previous iteration
    new_dataset: List[Artifact]  # training data used by current iteration
    previous: List[Artifact]  # previous models used by previous iteration


@dataclass
class CllDeepmdContext(BaseCllContext):
    config: CllDeepmdContextConfig


@dataclass
class GenericDeepmdOutput(ICllTrainOutput):
    models: List[Artifact]
    dataset: List[Artifact]

    def get_mlp_models(self) -> List[Artifact]:
        return self.models

    def get_training_dataset(self) -> List[Artifact]:
        return self.dataset


async def cll_deepmd(input: CllDeepmdInput, ctx: CllDeepmdContext):
    executor = ctx.resource_manager.default_executor

    # if fix_models is set, return it and skip training
    if len(input.config.fixture_models) > 0:
        logger.info(f'Using fixture models: {input.config.fixture_models}')
        model_paths = executor.run_python_fn(expand_globs)(input.config.fixture_models)
        assert len(model_paths) > 0, f'No fixture models found: {input.config.fixture_models}'
        return GenericDeepmdOutput(
            dataset=input.old_dataset.copy(),
            models=[Artifact.of(url=url, format=DataFormat.DEEPMD_MODEL)
                    for url in model_paths]
        )

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [new_dataset_dir, tasks_dir, outlier_dir] = executor.setup_workspace(
        work_dir, ['new_dataset', 'tasks', 'outlier_dataset'])

    init_dataset = ctx.resource_manager.resolve_artifacts(input.config.init_dataset)
    # input dataset contains data that generated by previous iteration
    # it should not contain the initial dataset
    input_dataset: List[Artifact] = input.old_dataset.copy()

    # make new dataset from raw data
    new_dataset, outlier_dataset = executor.run_python_fn(make_deepmd_dataset)(
        dataset_dir=new_dataset_dir,
        outlier_dir=outlier_dir,
        raw_data_collection=[a.to_dict() for a in input.new_dataset],
        type_map=input.type_map,
        isolate_outliers=input.config.isolate_outliers,
        outlier_f_cutoff=input.config.outlier_f_cutoff,
        group_by_formula=input.config.group_by_formula,
        deepmd_input_template=input.config.input_template,
        sel_type=input.sel_type,
        mode=input.mode,
        ignore_error=input.config.ignore_error,
    )

    input_dataset += [ Artifact.of(**a) for a in new_dataset]
    # use attrs to distinguish outlier dataset
    input_dataset += [ Artifact.of(**{**a, 'attrs': {**a['attrs'], 'outlier': True}}) for a in outlier_dataset]
    # classify dataset
    train_systems, outlier_systems, validation_systems = _classify_dataset(input_dataset + init_dataset)

    # setup deep wannier training job if needed
    dw_input_template = None
    if input.config.train_dw and input.mode == 'dpff':
        dw_input_template = input.config.train_dw.input_template

    # make task dirs
    dp_task_dirs, dw_task_dir = executor.run_python_fn(make_deepmd_task_dirs)(
        input_template=input.config.input_template,
        model_num=input.config.model_num,
        type_map=input.type_map,
        train_systems=train_systems,
        outlier_systems=outlier_systems,
        validation_systems=validation_systems,
        outlier_weight=input.config.outlier_weight,
        isolate_outliers=input.config.isolate_outliers,
        dw_input_template=dw_input_template,
        base_dir=tasks_dir,
    )

    # run dw training job if needed
    if dw_task_dir is not None:
        logger.info(f'Run deep wannier training job, output dir: {dw_task_dir}')
        steps = _build_deepmd_steps(
            dp_cmd=ctx.config.dp_cmd,
            compress_model=input.config.compress_model,
            cwd=dw_task_dir,
            pretrained_model=input.config.pretrained_model,
        )
        dw_train_script = BashScript(
            template=ctx.config.script_template,
            steps=steps,
        )
        job = executor.submit(dw_train_script.render(), cwd=tasks_dir)
        await gather_jobs([job], max_tries=2)
        logger.info(f'Deep wannier training job finished, output dir: {dw_task_dir}')

    all_steps: List[list] = []
    # run dp training jobs

    for i, task_dir in enumerate(dp_task_dirs):
        executor.mkdir(task_dir)
        # FIXME: a temporary workaround to support previous model
        previous_model = None
        if input.config.init_from_previous and input.previous:
            j = i % len(input.previous)
            previous_model = os.path.join(os.path.dirname(input.previous[j].url),
                                          DP_ORIGINAL_MODEL)

        steps = _build_deepmd_steps(
            dp_cmd=ctx.config.dp_cmd,
            compress_model=input.config.compress_model,
            cwd=task_dir,
            pretrained_model=input.config.pretrained_model,
            previous_model=previous_model,
        )
        all_steps.append(steps)

    # submit jobs by the number of concurrency
    jobs = []
    for i, steps_group in enumerate(list_split(all_steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=flatten(steps_group)
        )
        job = executor.submit(script.render(), cwd=tasks_dir)
        jobs.append(job)

    await gather_jobs(jobs, max_tries=2)

    logger.info(f'All models are trained, output dirs: {dp_task_dirs}')
    return GenericDeepmdOutput(
        dataset=input_dataset.copy(),
        models=[Artifact.of(
            url=os.path.join(url, DP_FROZEN_MODEL),
            format=DataFormat.DEEPMD_MODEL,
        ) for url in dp_task_dirs]
    )

def _classify_dataset(dataset: List[Artifact]):
    """
    Classify dataset into train, outlier and validation
    """
    train_systems: List[str] = []
    outlier_systems: List[str] = []
    validation_systems: List[str] = []

    for a in dataset:
        if dict_nested_get(a.attrs, ['deepmd', 'validation_data'], False):
            validation_systems.append(a.url)
        else:
            if dict_nested_get(a.attrs, ['outlier'], False):
                outlier_systems.append(a.url)
            else:
                train_systems.append(a.url)
    # FIXME: there are users reporting duplicated data in dataset
    # So we need to check and remove duplicated data
    # TODO: find the root cause of duplicated data and remove this workaround
    def _unique(l: list):
        r = sorted(set(l))
        if len(r) != len(l):
            logger.warning(f'Found duplicated data in dataset: {l}')
        return r
    return _unique(train_systems), _unique(outlier_systems), _unique(validation_systems)


def _build_deepmd_steps(dp_cmd: str,
                        compress_model: bool,
                        cwd: str,
                        previous_model: Optional[str] = None,
                        pretrained_model: Optional[str] = None,):
    steps = []
    dp_train_cmd = f'{dp_cmd} train {DP_INPUT_FILE}'

    if previous_model:
        dp_train_cmd = f'{dp_train_cmd} -f {previous_model}'
    if pretrained_model:
        dp_train_cmd = f'{dp_train_cmd} --finetune {pretrained_model}'

    dp_train_cmd_restart = f'if [ ! -f model.ckpt.index ]; then {dp_train_cmd}; else {dp_cmd} train {DP_INPUT_FILE} --restart model.ckpt; fi'

    steps.append(
        BashStep(cmd=dp_train_cmd_restart, cwd=cwd, checkpoint='dp-train')  # type: ignore
    )
    if compress_model:
        steps.append(BashStep(cmd=[dp_cmd, 'freeze', '-o', DP_ORIGINAL_MODEL, '&&',
                                   dp_cmd, 'compress', '-i', DP_ORIGINAL_MODEL, '-o', DP_FROZEN_MODEL],
                              cwd=cwd))
    else:
        # FIXME: a temporary workaround to support previous model
        steps.append(BashStep(cmd=[dp_cmd, 'freeze', '-o', DP_ORIGINAL_MODEL, '&&',
                                   'cp', DP_ORIGINAL_MODEL, DP_FROZEN_MODEL],
                              cwd=cwd))
    return steps


def make_deepmd_task_dirs(input_template: dict,
                            model_num: int,
                            type_map: List[str],
                            train_systems: List[str],
                            outlier_systems: List[str],
                            validation_systems: List[str],
                            isolate_outliers: bool,
                            outlier_weight: float,
                            dw_input_template: Optional[dict],
                            base_dir: str,
                            ):
    dp_task_dirs = [os.path.join(base_dir, f'{i:03d}')  for i in range(model_num)]
    for task_dir in dp_task_dirs:
        os.makedirs(task_dir, exist_ok=True)
        dp_input = make_deepmd_input(
            input_template=input_template,
            type_map=type_map,
            train_systems=train_systems,
            outlier_systems=outlier_systems,
            validation_systems=validation_systems,
            isolate_outliers=isolate_outliers,
            outlier_weight=outlier_weight,
        )

        # If dw_model is set in dp_input,
        # it will create a softlink named dw_model.pb to the current dir,
        # and then modify the value in dw_model to dw_model.pb.
        dw_model = dict_nested_get(dp_input, ['model', 'modifier', 'model_name'], None)
        if dw_model is not None:
            # Create a soft link named 'dw_model.pb' to the file specified by dw_model
            link_target = os.path.join(task_dir, 'dw_model.pb')
            os.system(f'ln -sf {dw_model} {link_target}')
            # Modify the value in dw_model to 'dw_model.pb'
            dp_input['model']['modifier']['model_name'] = 'dw_model.pb'

        dp_input_path = os.path.join(task_dir, DP_INPUT_FILE)
        dump_json(dp_input, dp_input_path)

    dw_task_dir = None
    if dw_input_template:
        dw_task_dir = os.path.join(base_dir, 'train_dw')
        os.makedirs(dw_task_dir, exist_ok=True)
        dw_input = make_deepmd_input(
            input_template=dw_input_template,
            type_map=type_map,
            train_systems=train_systems,
            outlier_systems=outlier_systems,
            validation_systems=validation_systems,
            isolate_outliers=isolate_outliers,
            outlier_weight=outlier_weight,
        )
        dw_input_path = os.path.join(dw_task_dir, DP_INPUT_FILE)
        dump_json(dw_input, dw_input_path)

    return dp_task_dirs, dw_task_dir


def make_deepmd_input(input_template: dict,
                        type_map: List[str],
                        train_systems: List[str],
                        outlier_systems: List[str],
                        validation_systems: List[str],
                        isolate_outliers: bool,
                        outlier_weight: float,
                        ):
    # create dp train input file
    # ref: https://github.com/deepmodeling/dpgen2/blob/master/examples/ch4/param_CH4_deepmd-kit-2.1.1.json
    # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/prep_dp_train.py
    # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/run_dp_train.py

    def _random_seed():
        return random.randrange(sys.maxsize) % (1 << 32)

    dp_input = copy.deepcopy(input_template)
    training: dict = dp_input['training']

    # set output files
    training['disp_file'] = DP_DISP_FILE
    training['save_ckpt'] = DP_CHECKPOINT_FILE
    training['profiling_file'] = DP_PROFILING_FILE

    # set random seed
    discriptor = dp_input['model']['descriptor']
    if discriptor['type'] == 'hybrid':
        for d in discriptor['list']:
            d['seed'] = _random_seed()
    else:
        discriptor['seed'] = _random_seed()
    dp_input['model']['fitting_net']['seed'] = _random_seed()
    dp_input['training']['seed'] = _random_seed()

    auto_prob_str = "prob_sys_size"
    if isolate_outliers and len(outlier_systems) > 0:
        # if isolate_outlier is enabled, then we need to set different weight for outlier data
        # e.g: prob_sys_size;0:57:0.9997;57:60:0.0003
        auto_prob_str += f';0:{len(train_systems)}:{1-outlier_weight}'
        auto_prob_str += f';{len(train_systems)}:{len(train_systems)+len(outlier_systems)}:{outlier_weight}'

    # v2 training data
    training_data = training.get('training_data', {})
    training_data = {
        'set_prefix': 'set',
        'batch_size': 'auto',
        'sys_probs': None,
        ** training_data,
        'systems': train_systems + outlier_systems,
        'auto_prob_style': auto_prob_str,
    }
    training['training_data'] = training_data

    # ignore validation section if no data is provided, or else dp will throw error
    # OSError: [Errno cannot find valid a data system] Please check your setting for data systems
    if len(validation_systems) > 0:
        validation_data = {
            'systems': validation_systems,
            'set_prefix': training_data['set_prefix'],
            'batch_size': training_data['batch_size'],
        }
        training['validation_data'] = validation_data

    # other params
    dp_input['model']['type_map'] = type_map

    return dp_input

def make_deepmd_dataset(
    dataset_dir: str,
    outlier_dir: str,
    raw_data_collection: List[ArtifactDict],
    isolate_outliers: bool,
    outlier_f_cutoff: float,
    type_map: List[str],
    deepmd_input_template: dict,
    group_by_formula: bool,
    mode: str,
    sel_type: Optional[List[int]],
    ignore_error: bool,
):
    register_data_types()

    dataset_collection: List[Tuple[ArtifactDict, dpdata.LabeledSystem]] = []
    outlier_collection: List[Tuple[ArtifactDict, dpdata.LabeledSystem]] = []

    for raw_data in raw_data_collection:
        data_format = get_data_format(raw_data)  # type: ignore
        dp_system = None
        try:
            if data_format == DataFormat.CP2K_OUTPUT_DIR:
                dp_system = dpdata.LabeledSystem(os.path.join(raw_data['url'], 'output'), fmt='cp2k/output', type_map=type_map)
            elif data_format == DataFormat.VASP_OUTPUT_DIR:
                dp_system = dpdata.LabeledSystem(os.path.join(raw_data['url'], 'OUTCAR'), fmt='vasp/outcar', type_map=type_map)
            else:
                raise ValueError(f"Unsupported data format: {data_format}")
        except Exception as e:
            if not ignore_error:
                raise e
            logger.exception(f'Failed to load data: {raw_data}, error: {e}')
        # one case of len(dp_system) == 0 is when the system is not converged
        if dp_system is None or 0 == len(dp_system):
            continue  # skip invalid data

        # generate extra data for dpff
        if mode == 'dpff':
            modifier = deepmd_input_template['model']['modifier']
            assert sel_type is not None, 'sel_type must be set in dpff mode'

            if data_format == DataFormat.CP2K_OUTPUT_DIR:
                # Arguments of DPLR model can be found here:
                # https://github.com/deepmodeling/deepmd-kit/blob/master/doc/model/dplr.md
                try:
                    set_dplr_ext_from_cp2k_output(
                        dp_sys=dp_system,
                        cp2k_output=os.path.join(raw_data['url'], 'output'),
                        wannier_file=os.path.join(raw_data['url'], 'wannier.xyz'),
                        ext_efield=raw_data['attrs']['efield'],
                        type_map=type_map,
                        sys_charge_map=modifier['sys_charge_map'],
                        model_charge_map=modifier['model_charge_map'],
                        ewald_h=modifier['ewald_h'],
                        ewald_beta=modifier['ewald_beta'],
                        sel_type=sel_type,
                    )
                except Exception:
                    logger.exception(f'dpff: failed to set dplr ext')
                    continue
            else:
                raise ValueError(f"Unsupported data format: {data_format}")

        # set fparam if existed
        fparam = raw_data['attrs'].get('dp_fparam', None)
        if fparam is not None:
            set_fparam(dp_system, fparam)

        if isolate_outliers and dp_system.data['forces'].max() > outlier_f_cutoff:
            outlier_collection.append((raw_data, dp_system))
        else:
            dataset_collection.append((raw_data, dp_system))

    _write_dp_dataset = _write_dp_dataset_by_formula if group_by_formula else _write_dp_dataset_by_ancestor

    dataset_dirs = _write_dp_dataset(dp_system_list=dataset_collection, out_dir=dataset_dir, type_map=type_map)
    outlier_dirs = _write_dp_dataset(dp_system_list=outlier_collection, out_dir=outlier_dir, type_map=type_map)

    return dataset_dirs, outlier_dirs


def _write_dp_dataset_by_formula(dp_system_list: List[Tuple[ArtifactDict, dpdata.LabeledSystem]],
                                    out_dir: str, type_map: List[str]):
    """
    Write dp dataset that grouping by formula
    Use dpdata.MultipleSystems to merge systems with the same formula
    Use this when group by ancestor not works for you
    """
    if len(dp_system_list) == 0:
        return []
    os.makedirs(out_dir, exist_ok=True)

    multi_systems = dpdata.MultiSystems(dp_system_list[0][1])
    for _, system in dp_system_list[1:]:
        multi_systems.append(system)

    # pylint: disable=no-member
    multi_systems.to_deepmd_npy(out_dir, type_map=type_map)  # type: ignore

    return [ {
        'url': os.path.join(out_dir, fname),
        'format': DataFormat.DEEPMD_NPY,
        'attrs': {},  # it is meaning less to set attrs in this case
    } for fname in os.listdir(out_dir)]


def _write_dp_dataset_by_ancestor(dp_system_list: List[Tuple[ArtifactDict, dpdata.LabeledSystem]], out_dir: str, type_map: List[str]):
    """
    write dp dataset that grouping by ancestor
    """
    output_dirs: List[ArtifactDict] = []
    def get_ancestor(x):
        return x[0]['attrs']['ancestor']
    dp_system_list = sorted(dp_system_list, key=get_ancestor)
    for key, dp_system_group in groupby(dp_system_list, key=get_ancestor):
        dp_system_group = list(dp_system_group)
        if 0 == len(dp_system_group):
            continue  # skip empty dataset
        group_out_dir = os.path.join(out_dir, key.replace('/', '_'))

        # merge dp_systems with the same ancestor into single data set
        dp_system = dp_system_group[0][1]
        for item in dp_system_group[1:]:
            dp_system += item[1]
        dp_system.to_deepmd_npy(group_out_dir, set_size=len(dp_system), type_map=type_map)  # type: ignore
        # inherit attrs key from input artifact
        output_dirs.append({'url': group_out_dir,
                            'format': DataFormat.DEEPMD_NPY,
                            'attrs': {**dp_system_group[0][0]['attrs']}})  # type: ignore
    return output_dirs



import logging
import os

level_name = os.environ.get('LOG_LEVEL', 'INFO')
level = logging._nameToLevel.get(level_name, logging.INFO)

# format to include timestamp and module
logging.basicConfig(format='%(asctime)s %(name)s: %(message)s', level=level)
# suppress transitions logging
logging.getLogger('transitions.core').setLevel(logging.WARNING)

def get_logger(name=None):
    return logging.getLogger(name)


from ai2_kit import res
import os

AI2CAT_RES_DIR = os.path.join(res.DIR_PATH, 'catalysis')

DEEPMD_DEFAULT_TEMPLATE = os.path.join(AI2CAT_RES_DIR, 'deepmd.json')
MLP_TRAINING_TEMPLATE   = os.path.join(AI2CAT_RES_DIR, 'mlp-training.yml')
CP2K_DEFAULT_TEMPLATE   = os.path.join(AI2CAT_RES_DIR, 'cp2k.inp')

CP2K_SCF_SEMICONDUCTOR = """\
        # CONFIGURATION FOR SEMICONDUCTOR
        &SCF
            SCF_GUESS RESTART
            EPS_SCF 3e-07
            MAX_SCF 50
            &OUTER_SCF
                EPS_SCF 3e-07
                MAX_SCF 20
            &END OUTER_SCF
            &OT
                MINIMIZER CG
                PRECONDITIONER FULL_SINGLE_INVERSE
                ENERGY_GAP 0.1
            &END OT
        &END SCF
        # END CONFIGURATION FOR SEMICONDUCTOR"""

CP2K_SCF_METAL = """\
        # CONFIGURATION FOR METAL
        &SCF
            SCF_GUESS RESTART
            EPS_SCF 3e-07
            MAX_SCF 500
            ADDED_MOS 500
            CHOLESKY INVERSE
            &SMEAR
                METHOD FERMI_DIRAC
                ELECTRONIC_TEMPERATURE [K] 300
            &END SMEAR
            &DIAGONALIZATION
                ALGORITHM STANDARD
            &END DIAGONALIZATION
            &MIXING
                METHOD BROYDEN_MIXING
                ALPHA 0.3
                BETA 1.5
                NBROYDEN 14
            &END MIXING
        &END SCF
        # END CONFIGURATION FOR METAL"""

CP2K_MOTION_TEMPLATE = """\
&MOTION
  &MD
    ENSEMBLE NVT
    STEPS       $steps
    TIMESTEP    $timestep
    TEMPERATURE $temp
    &THERMOSTAT
       TYPE CSVR
       REGION MASSIVE
       &CSVR
          TIMECON [fs] 100.0
       &END
    &END
  &END MD
  &PRINT
   &TRAJECTORY
     &EACH
       MD 1
     &END EACH
   &END TRAJECTORY
   &VELOCITIES
     &EACH
       MD 1
     &END EACH
   &END VELOCITIES
   &FORCES
     &EACH
       MD 1
     &END EACH
   &END FORCES
   &RESTART_HISTORY
     &EACH
       MD 1000
     &END EACH
   &END RESTART_HISTORY
   &RESTART
     BACKUP_COPIES 3
     &EACH
       MD 1
     &END EACH
   &END RESTART
  &END PRINT
&END MOTION"""

CP2K_SCF_TABLE = {
    'metal': CP2K_SCF_METAL,
    'semi': CP2K_SCF_SEMICONDUCTOR,
}

CP2K_ACCURACY_TABLE = {
    'high': {'cutoff': 1000, 'rel_cutoff': 90 },
    'medium': {'cutoff': 800, 'rel_cutoff': 70 },
    'low': {'cutoff': 600, 'rel_cutoff': 50 },
}


from typing import List, Dict, Optional
from collections import OrderedDict
from ase import Atoms
import numpy as np
import ase.io
import dpdata
import shlex
import os

from MDAnalysis.lib.distances import distance_array, minimize_vectors

from ai2_kit.core.util import list_split, expand_globs, cmd_with_checkpoint
from ai2_kit.domain.cp2k import dump_coord_n_cell
from ai2_kit.core.log import get_logger

logger = get_logger(__name__)


class Cp2kLabelTaskBuilder:
    """
    This a builder class to generate cp2k labeling task.
    """
    def __init__(self):
        self._atoms_list: List[Atoms] = []
        self._task_dirs: List[str] = []
        self._cp2k_inputs: OrderedDict = OrderedDict()

    def add_system(self, *file_path_or_glob: str, **kwargs):
        """
        Add system files to label

        :param file_path_or_glob: system files or glob pattern
        :param kwargs: kwargs for ase.io.read
        """
        files = expand_globs(file_path_or_glob)
        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._ase_read(file, **kwargs)
        return self

    def add_cp2k_input(self, file_path: str, tag: str):
        """
        Add cp2k input file for dipole or polarizability calculation
        """
        assert tag not in self._cp2k_inputs, f'Tag {tag} already exists'
        self._cp2k_inputs[tag] = _ensure_file_exists(file_path)
        return self

    def make_tasks(self, out_dir: str):
        """
        Make task dirs for cp2k labeling (prepare systems and cp2k input files)
        :param out_dir: output dir
        """
        assert self._atoms_list, 'No system files added'
        assert not self._task_dirs, 'Task dirs already generated'

        task_dirs = []
        for i, atoms in enumerate(self._atoms_list):
            task_name = f'{i:06d}'
            task_dir = os.path.join(out_dir, task_name)
            os.makedirs(task_dir, exist_ok=True)
            # dump coord_n_cell.inc, so that we can use @include to read system
            cp2k_sys_file = os.path.join(task_dir, 'coord_n_cell.inc')
            with open(cp2k_sys_file, 'w') as f:
                dump_coord_n_cell(f, atoms)
            # dump system.xyz for cp2k,
            # so that use can use COORD_FILE_NAME to read system
            xyz_sys_file = os.path.join(task_dir, 'system.xyz')
            ase.io.write(xyz_sys_file, atoms, format='extxyz', append=True)
            # link cp2k input files to tasks dirs
            for tag, file_path in self._cp2k_inputs.items():
                link_path = os.path.join(task_dir, tag + '.inp')
                os.system(f'ln -sf {os.path.abspath(file_path)} {link_path}')
            task_dirs.append(task_dir)
        self._task_dirs = task_dirs
        logger.info(f'Generated {len(task_dirs)} task dirs')
        return self

    def make_scripts(self,
                     prefix: str = 'cp2k-batch-{i:02d}.sub',
                     concurrency: int = 5,
                     template: Optional[str] = None,
                     ignore_error: bool = False,
                     cp2k_cmd: str = 'cp2k.psmp'):
        """
        Make batch script for cp2k labeling

        :param prefix: prefix of batch script
        :param concurrency: concurrency of tasks
        :param template: template of batch script
        :param ignore_error: ignore error
        :param cmd: command to run cp2k
        """
        assert self._atoms_list, 'No system files added, please call add_system first'
        assert self._task_dirs, 'No task dirs generated, please call make_dirs first'
        if template is None:
            template = '#!/bin/bash'
        else:
            with open(template, 'r') as f:
                template = f.read()
        template += '\n[ -n "$PBS_O_WORKDIR" ] && cd $PBS_O_WORKDIR]\n'

        os.makedirs(os.path.dirname(prefix), exist_ok=True)
        for i, task_group in enumerate(list_split(self._task_dirs, concurrency)):
            if not task_group:
                continue
            batch = [template]
            batch.append("for _WORK_DIR_ in \\")
            # use shell for loop to run tasks
            for task_dir in task_group:
                batch.append(f'  {shlex.quote(task_dir)} \\')
            batch.extend([
                f'; do',
                f'pushd $_WORK_DIR_ || exit 1',
                '#' * 80,
                *[
                    cmd_with_checkpoint(f'{cp2k_cmd} -i {tag}.inp &> {tag}.out', f'{tag}.done', ignore_error)
                    for tag in self._cp2k_inputs.keys()
                ],
                '#' * 80,
                f'popd',
                f'done'
            ])
            batch_file = prefix.format(i=i)
            with open(batch_file, 'w', encoding='utf-8') as f:
                f.write('\n'.join(batch))
            logger.info(f'Write batch script to {batch_file}')
        return self

    def done(self):
        """
        Mark the end of commands chain
        """

    def _ase_read(self, filename: str, **kwargs):
        kwargs.setdefault('index', ':')
        data = ase.io.read(filename, **kwargs)
        if not isinstance(data, list):
            data = [data]
        self._atoms_list += data
        logger.info(f'Loaded {len(data)} systems from {filename}, total {len(self._atoms_list)} systems')


def dpdata_read_cp2k_viber_data(data_dir: str,
                                lumped_dict: Dict[str, int],
                                output_file: str = 'output',
                                wannier: str = 'wannier.xyz',
                                wannier_x: str = 'wannier_x.xyz',
                                wannier_y: str = 'wannier_y.xyz',
                                wannier_z: str = 'wannier_z.xyz',
                                wacent_symbol='X',
                                cutoff = 1.2,
                                eps = 1e-3,
                                mode = 'both'
                                ):

    """
    read the cp2k output file and wannier functions

    :param data_dir: the directory of the data
    :param lumped_dict: the dictionary of the element and the expected coordination number, e.g. {"O": 4} (for water molecule)
    :param output_file: the cp2k output file name
    :param wannier: the wannier function file name
    :param wannier_x: the wannier function file name of x direction electric field
    :param wannier_y: the wannier function file name of y direction electric field
    :param wannier_z: the wannier function file name of z direction electric field
    :param wacent_symbol: the symbol of the wannier function centroid
    """

    dp_sys = dpdata.LabeledSystem(os.path.join(data_dir, output_file) , fmt='cp2k/output')

    # get cell
    cell = dp_sys.data['cells'][0]

    # build the data of atomic_dipole and atomic_polarizability with numpy
    wannier_atoms = ase.io.read(os.path.join(data_dir, wannier), index=":", format='extxyz')

    lumped_dict_c = lumped_dict.copy()
    del_list = []
    for k in lumped_dict_c.keys():
        if k not in wannier_atoms[0].get_chemical_symbols():
            del_list.append(k)

    for i in del_list:
        del lumped_dict_c[i]

    stc_list = _set_cells(wannier_atoms, cell)  # type: ignore
    wfc_compute_polar = _set_lumped_wfc(stc_list, lumped_dict_c, cutoff, wacent_symbol, to_polar = True)
    wfc_save = _set_lumped_wfc(stc_list, lumped_dict_c, cutoff, wacent_symbol, to_polar = False)

    dp_sys.data['atomic_dipole'] = wfc_save

    if mode == 'both':
        wannier_atoms_x = ase.io.read(os.path.join(data_dir, wannier_x), index=":", format='extxyz')
        stc_list = _set_cells(wannier_atoms_x, cell)  # type: ignore
        wfc_x = _set_lumped_wfc(stc_list, lumped_dict_c, cutoff, wacent_symbol, to_polar = True)

        wannier_atoms_y = ase.io.read(os.path.join(data_dir, wannier_y), index=":", format='extxyz')
        stc_list = _set_cells(wannier_atoms_y, cell)  # type: ignore
        wfc_y = _set_lumped_wfc(stc_list, lumped_dict_c, cutoff, wacent_symbol, to_polar = True)

        wannier_atoms_z = ase.io.read(os.path.join(data_dir, wannier_z), index=":", format='extxyz')
        stc_list = _set_cells(wannier_atoms_z, cell)  # type: ignore
        wfc_z = _set_lumped_wfc(stc_list, lumped_dict_c, cutoff, wacent_symbol, to_polar = True)

        polar = np.zeros((wfc_compute_polar.shape[0], wfc_compute_polar.shape[1], 3), dtype = float)

        polar[:, :, 0] = (wfc_x - wfc_compute_polar) / eps
        polar[:, :, 1] = (wfc_y - wfc_compute_polar) / eps
        polar[:, :, 2] = (wfc_z - wfc_compute_polar) / eps

        dp_sys.data['atomic_polarizability'] = polar.reshape(polar.shape[0], -1)
    elif mode == 'dipole_only':
        dp_sys.data['atomic_polarizability'] = np.array([-1,-1]).reshape(1,-1)
    else:
        logger.warning(f"There is no mode called '{mode}', expected 'both' or 'dipole_only'")

    return dp_sys


def _get_lumped_wacent_poses_rel(stc: Atoms, elem_symbol, wacent_symbol, cutoff=1.2, expected_cn=4):
    """
    determine the positions of the wannier centers around O
    and sum it into the wannier centroid.
    """
    elem_idx = np.where(stc.symbols == elem_symbol)[0]
    wacent_idx = np.where(stc.symbols == wacent_symbol)[0]
    elem_poses = stc.positions[elem_idx]
    wacent_poses = stc.positions[wacent_idx]

    cellpar = stc.cell.cellpar()
    assert cellpar is not None, "cellpar is None"
    #dist_matrix
    dist_mat = distance_array(elem_poses, wacent_poses, box=cellpar)

    #each row get distance and select the candidates
    lumped_wacent_poses_rel = []
    for elem_entry, dist_vec in enumerate(dist_mat):
        bool_vec = (dist_vec < cutoff)
        cn = np.sum(bool_vec)

        # modify neighbor wannier centers coords relative to the center element atom
        neig_wacent_poses = wacent_poses[bool_vec, :]
        neig_wacent_poses_rel = neig_wacent_poses - elem_poses[elem_entry]
        neig_wacent_poses_rel = minimize_vectors(neig_wacent_poses_rel, box=cellpar)
        lumped_wacent_pos_rel = neig_wacent_poses_rel.mean(axis=0)

        if cn != expected_cn:
            logger.warning(f"Coordination number of {elem_symbol} is {cn}, expected {expected_cn}")

        lumped_wacent_poses_rel.append(lumped_wacent_pos_rel)
    lumped_wacent_poses_rel = np.stack(lumped_wacent_poses_rel)
    return lumped_wacent_poses_rel

def _can_retain(item,elem_symbol):
    if not isinstance(item,str): return True
    if item != elem_symbol: return True
    return False

def _is_X(item):
    if not isinstance(item,str): return False
    if item == 'X': return True
    return False

def _set_lumped_wfc(stc_list, lumped_dict, cutoff, wacent_symbol, to_polar = True):
    """
    set the wannier function centroids
    """
    X_pos = []
    if to_polar:
        for stc in stc_list:
            x_symbol = list(stc.symbols)
            
            for elem_symbol, expected_cn in lumped_dict.items():
                lumped_wacent_poses_rel = _get_lumped_wacent_poses_rel(
                    stc=stc, elem_symbol=elem_symbol, wacent_symbol = wacent_symbol,
                    cutoff = cutoff, expected_cn=expected_cn)
                out_elem_symbol = list(lumped_wacent_poses_rel)
                x_symbol = [item if _can_retain(item,elem_symbol) else out_elem_symbol.pop(0) for item in x_symbol]

            x_symbol = [item for item in x_symbol if not _is_X(item)]
            x_symbol = [np.array([0.,0.,0.]) if isinstance(item, str) else item for item in x_symbol]
            x_symbol = np.concatenate(x_symbol ,axis = 0)
            X_pos.append(np.array(x_symbol))
        wfc_pos = np.array(X_pos)
        return wfc_pos
    else:
        for stc in stc_list:
            for elem_symbol, expected_cn in lumped_dict.items():
                lumped_wacent_poses_rel = _get_lumped_wacent_poses_rel(
                    stc=stc, elem_symbol=elem_symbol, wacent_symbol = wacent_symbol,
                    cutoff = cutoff, expected_cn=expected_cn)
                X_pos.append(np.reshape(lumped_wacent_poses_rel, (len(stc_list), -1)))
        wfc_pos = np.concatenate(X_pos,axis = 1)
        return wfc_pos

def _set_cells(stc_list: List[Atoms], cell):
    for stc in stc_list:
        stc.set_cell(cell)
        stc.set_pbc(True)
    return stc_list


def _ensure_file_exists(file_path: str):
    assert os.path.isfile(file_path), f'{file_path} is not a file'
    return file_path


cmd_entry = {
    'cp2k-labeling': Cp2kLabelTaskBuilder,
}



from ai2_kit.core.util import expand_globs
from ai2_kit.core.log import get_logger
from ai2_kit.core.job import JobState
from ai2_kit.core.cmd import CmdGroup

import datetime
import shlex
import time
import os
import re


logger = get_logger(__name__)


def append_if_not_exist(fp, line: str, feat_str = None):
    if feat_str is None:
        feat_str = line
    for line in fp:
        if feat_str in line:
            return
    fp.seek(0, os.SEEK_END)
    fp.write(f'\n{line}\n')


class Slurm:

    _state_table = {
        'PD': JobState.PENDING,
        'R': JobState.RUNNING,
        'CA': JobState.CANCELLED,
        'CF': JobState.PENDING,  # (configuring),
        'CG': JobState.RUNNING,  # (completing),
        'CD': JobState.COMPLETED,
        'F': JobState.FAILED,  # (failed),
        'TO': JobState.TIMEOUT,  # (timeout),
        'NF': JobState.FAILED,  # (node failure),
        'RV': JobState.FAILED,  # (revoked) and
        'SE': JobState.FAILED   # (special exit state)
    }

    def __init__(self, sbatch_bin = 'sbatch', squeue_bin = 'squeue', scancel_bin='scancel') -> None:
        self._sbatch_bin = sbatch_bin
        self._squeue_bin = squeue_bin
        self._scancel_bin = scancel_bin
        self._job_states = {}

    def submit(self, *path_or_glob: str):
        """
        Submit multiple Slurm script at once.

        :param path_or_glob: path or glob of Slurm script
        """
        files = expand_globs(path_or_glob, raise_invalid=True)
        if not files:
            raise ValueError('No files found')
        for file in files:
            with open(file, '+a') as fp:
                append_if_not_exist(fp, 'touch slurm_$SLURM_JOB_ID.done # AUTO GENERATED')
        try:
            for file in files:
                job_id = self._submit(file)
                logger.info(f'Submitted job {job_id} for {file}')
                self._job_states[job_id] = JobState.UNKNOWN
        except:
            self._cancel()
            raise
        return self

    def wait(self, timeout=3600 * 24 * 7, ignore_error = False, fast_fail = False, interval = 10):
        """
        Wait until all jobs are finished.

        :param timeout: timeout in seconds
        :param ignore_error: ignore error
        :param fast_fail: exit if any job failed
        :param interval: interval in seconds
        """
        fail_cnt = 0
        start_at = datetime.datetime.now()
        while (datetime.datetime.now() - start_at).total_seconds() < timeout:
            try:
                self._update_job_states()
                logger.info('Job states: %s', self._job_states)
                fail_cnt = 0
            except Exception:
                fail_cnt += 1
                logger.exception('Failed to update job states')
                if fail_cnt > 5:  # stop if keep failing
                    raise
            if all(state.terminal for state in self._job_states.values()):
                break
            if fast_fail and self._is_any_failed():
                self._cancel()
                raise RuntimeError('Fast fail!')
            time.sleep(interval)
        else:
            logger.error('Timeout')
            if not ignore_error:
                raise RuntimeError('Timeout')
        if not ignore_error and self._is_any_failed():
            raise RuntimeError('Some jobs failed')

    def _is_any_failed(self):
        return any(state == JobState.FAILED for state in self._job_states.values())

    def _update_job_states(self):
        query_cmd = f"{self._squeue_bin} --noheader --format='%i %t' -u $USER"
        
        fp = os.popen(query_cmd)
        out = fp.read()
        exit_code = fp.close()
        if exit_code is not None:
            raise RuntimeError(f'Failed to query job states: {exit_code}')

        state = {}
        for line in out.splitlines():
            if line:
                job_id, slurm_state = line.split()
                state[job_id] = slurm_state
        for job_id in self._job_states:
            if job_id in state:
                self._job_states[job_id] = self._state_table[state[job_id]]
            else:
                if os.path.exists(f'slurm_{job_id}.done'):
                    self._job_states[job_id] = JobState.COMPLETED
                else:
                    self._job_states[job_id] = JobState.FAILED

    def _cancel(self):
        for job_id in self._job_states:
            os.system(f'{self._scancel_bin} {job_id}')

    def _submit(self, file: str):
        submit_cmd = f'{self._sbatch_bin} {shlex.quote(file)}'
        with os.popen(submit_cmd) as fp:
            stdout = fp.read()
        m = re.search(r'Submitted batch job (\d+)', stdout)
        if m is None:
            raise ValueError(f'Failed to submit job: {stdout}')
        job_id = m.group(1)
        return job_id

cmd_entry = CmdGroup(items={
    'slurm': Slurm,
}, doc='Tools to facilitate HPC related tasks.')



from typing import Tuple, List, TypeVar, Union, Iterable, Literal
from ruamel.yaml import YAML, ScalarNode, SequenceNode
from collections import namedtuple
from itertools import zip_longest
from pathlib import Path

import shortuuid
import asyncio
import hashlib
import inspect
import base64
import psutil
import random
import shlex
import json
import glob
import os

from .log import get_logger

logger = get_logger(__name__)

EMPTY = object()

SAMPLE_METHOD = Literal['even', 'random', 'truncate']

def load_json(path: Union[Path, str], encoding: str = 'utf-8'):
    if isinstance(path, str):
        path = Path(path)
    with open(path, 'r', encoding=encoding) as f:
        return json.load(f)


def load_text(path: Union[Path, str], encoding: str = 'utf-8'):
    if isinstance(path, str):
        path = Path(path)
    with open(path, 'r', encoding=encoding) as f:
        return f.read()


def parse_path_list(path_list_str: Union[str, List[str]], to_abs: bool = False):
    """
    Parse path list of environment variable style string
    """
    def parse_path(path: str):
        return os.path.expanduser(path) if path.startswith('~/') else path
    if isinstance(path_list_str, str):
        path_list = path_list_str.split(':')
    else:
        path_list = path_list_str
    if to_abs:
        path_list = [parse_path(path) for path in path_list]
    return path_list


def wait_for_change(widget, attribute):
    """
    Wait for attribute change of a Jupyter widget
    """
    future = asyncio.Future()
    def getvalue(change):
        # make the new value available
        future.set_result(change.new)
        widget.unobserve(getvalue, attribute)
    widget.observe(getvalue, attribute)
    return future


def get_yaml():
    yaml = YAML(typ='safe')
    JoinTag.register(yaml)
    LoadTextTag.register(yaml)
    LoadYamlTag.register(yaml)
    return yaml


def load_yaml_file(path: Union[Path, str]):
    if isinstance(path, str):
        path = Path(path)
    yaml = get_yaml()
    return yaml.load(path)


def load_yaml_files(*paths: Union[Path, str], quiet: bool = False, purge_anonymous = True):
    d = {}
    for path in paths:
        print('load yaml file: ', path)
        d = merge_dict(d, load_yaml_file(path), quiet=quiet)
    if purge_anonymous:
        dict_remove_dot_keys(d)
    return d


def nested_set(d: dict, keys: List[str], value):
    for key in keys[:-1]:
        d = d.setdefault(key, {})
    d[keys[-1]] = value


def s_uuid():
    """short uuid"""
    return shortuuid.uuid()


def sort_unique_str_list(l: List[str]) -> List[str]:
    """remove duplicate str and sort"""
    return sorted(set(l))


T = TypeVar('T')


def flatten(l: List[List[T]]) -> List[T]:
    return [item for sublist in l for item in sublist]


def format_env_string(s: str) -> str:
    return s.format(**os.environ)


def list_split(l: List[T], n: int) -> List[List[T]]:
    """split list into n chunks"""
    # ref: https://stackoverflow.com/questions/2130016/splitting-a-list-into-n-parts-of-approximately-equal-length
    k, m = divmod(len(l), n)
    return [l[i*k+min(i, m): (i+1)*k+min(i+1, m)] for i in range(n)]


def short_hash(s: str) -> str:
    """short hash string"""
    digest = hashlib.sha1(s.encode('utf-8')).digest()
    # use urlsafe encode to avoid '/' in the string, as it will cause problem in file path
    return base64.urlsafe_b64encode(digest).decode('utf-8')[:-2]


async def to_awaitable(value: T) -> T:
    return value


class JoinTag:
    """a tag to join strings in a list"""

    yaml_tag = u'!join'

    @classmethod
    def from_yaml(cls, constructor, node):
        seq = constructor.construct_sequence(node)
        return ''.join([str(i) for i in seq])

    @classmethod
    def to_yaml(cls, dumper, data):
        ...

    @classmethod
    def register(cls, yaml: YAML):
        yaml.register_class(cls)

class LoadTextTag:
    """a tag to read string from file"""

    yaml_tag = u'!load_text'

    @classmethod
    def from_yaml(cls, constructor, node):
        path = _yaml_get_path_node(node, constructor)
        with open(path, 'r') as f:
            return f.read()

    @classmethod
    def to_yaml(cls, dumper, data):
        ...

    @classmethod
    def register(cls, yaml: YAML):
        yaml.register_class(cls)


class LoadYamlTag:
    """a tag to read string from file"""

    yaml_tag = u'!load_yaml'

    @classmethod
    def from_yaml(cls, constructor, node):
        path = _yaml_get_path_node(node, constructor)
        yaml = get_yaml()
        with open(path, 'r') as f:
            return yaml.load(f)

    @classmethod
    def to_yaml(cls, dumper, data):
        ...

    @classmethod
    def register(cls, yaml: YAML):
        yaml.register_class(cls)


def _yaml_get_path_node(node, constructor):
    if isinstance(node, ScalarNode):
        return constructor.construct_scalar(node)
    elif isinstance(node, SequenceNode):
        seq = constructor.construct_sequence(node)
        return os.path.join(*seq)
    else:
        raise ValueError(f'Unknown node type {type(node)}')


def dict_remove_dot_keys(d):
    for k in list(d.keys()):
        if k.startswith('.'):
            del d[k]
        elif isinstance(d[k], dict):
            dict_remove_dot_keys(d[k])


def cmd_with_checkpoint(cmd: str, checkpoint: str, ignore_error: bool = False):
    """
    Add checkpoint to a shell command.
    If a checkpoint file exists, the command will be skipped,
    otherwise the command will be executed and
    a checkpoint file will be created if the command is executed successfully.

    :param cmd: shell command
    :param checkpoint: checkpoint file name
    :param ignore_error: if True, will not raise error if the command failed
    """
    checkpoint = shlex.quote(checkpoint)
    exit_clause = [
        '  __EXITCODE__=$?; if [ $__EXITCODE__ -ne 0 ]; then exit $__EXITCODE__; fi',
    ] if not ignore_error else []

    msg = shlex.quote(f"hit checkpoint: {checkpoint}, skip...")
    return '\n'.join([
        f'if [ -f {checkpoint} ]; then echo {msg}; else',
        f'  {cmd}',
        *exit_clause,
        f'  touch {checkpoint}',
        f'fi',
    ])


def merge_dict(lo: dict, ro: dict, path=None, ignore_none=True, quiet=False):
    """
    Merge two dict, the left dict will be overridden.
    Note: list will be replaced instead of merged.
    """
    if path is None:
        path = []
    for key, value in ro.items():
        if ignore_none and value is None:
            continue
        if key in lo:
            current_path = path + [str(key)]
            if isinstance(lo[key], dict) and isinstance(value, dict):
                merge_dict(lo[key], value, path=current_path, ignore_none=ignore_none, quiet=quiet)
            else:
                if not quiet:
                    print('.'.join(current_path) + ' has been overridden')
                lo[key] = value
        else:
            lo[key] = value
    return lo

def dict_nested_get(d: dict, keys: List[str], default=EMPTY):
    """get value from nested dict"""
    for key in keys:
        if key not in d and default is not EMPTY:
            return default
        d = d[key]
    return d

def dict_nested_set(d: dict, keys: List[str], value):
    """set value to nested dict"""
    for key in keys[:-1]:
        d = d[key]
    d[keys[-1]] = value

def list_even_sample(l, size):
    if size <= 0 or size > len(l):
        return l
    # calculate the sample interval
    interval = len(l) / size
    return [l[int(i * interval)] for i in range(size)]

def list_random_sample(l, size, seed = None):
    if seed is None:
        seed = len(l)
    random.seed(seed)
    return random.sample(l, size)

def list_sample(l, size: int, method: SAMPLE_METHOD='even', **kwargs):
    if method == 'even':
        return list_even_sample(l, size)
    elif method == 'random':
        return list_random_sample(l, size, **kwargs)
    elif method == 'truncate':
        return l[:size]
    else:
        raise ValueError(f'Unknown sample method {method}')

def flat_evenly(list_of_lists):
    """
    flat a list of lists and ensure the output result distributed evenly
    >>> flat_evenly([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    [1, 4, 7, 2, 5, 8, 3, 6, 9]
    Ref: https://stackoverflow.com/questions/76751171/how-to-flat-a-list-of-lists-and-ensure-the-output-result-distributed-evenly-in-p
    """
    return [e for tup in zip_longest(*list_of_lists) for e in tup if e is not None]


def limit(it, size=-1):
    """
    limit the size of an iterable
    """
    if size <= 0:
        yield from it
    else:
        for i, x in enumerate(it):
            if i >= size:
                break
            yield x


def dump_json(obj, path: str):
    default = lambda o: f"<<non-serializable: {type(o).__qualname__}>>"
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(obj, f, indent=2, default=default)


def dump_text(text: str, path: str, **kwargs):
    with open(path, 'w', **kwargs) as f:
        f.write(text)


def flush_stdio():
    import sys
    sys.stdout.flush()
    sys.stderr.flush()


def ensure_dir(path: str):
    dirname = os.path.dirname(path)
    if dirname:
        os.makedirs(dirname, exist_ok=True)


def expand_globs(patterns: Iterable[str], raise_invalid=False) -> List[str]:
    """
    Expand glob patterns in paths

    :param patterns: list of paths or glob patterns
    :param raise_invalid: if True, will raise error if no file found for a glob pattern
    :return: list of expanded paths
    """
    paths = []
    for pattern in patterns:
        result = glob.glob(pattern, recursive=True) if '*' in pattern else [pattern]
        if raise_invalid and len(result) == 0:
            raise FileNotFoundError(f'No file found for {pattern}')
        paths += result
    return sort_unique_str_list(paths)


def slice_from_str(index: str):
    """
    get slice object from a string expression,
    for example: "1:10" -> slice(1, 10), ":10" -> slice(None, 10), etc
    """
    parse = lambda s: int(s) if s else None
    parts = index.split(':')
    return slice(*(parse(s) for s in parts))


def perf_log(msg):
    process = psutil.Process()
    mem_info = process.memory_info()
    logger.info('perf: rss: %s M %s', mem_info.rss / 1024 ** 2, msg)


from dpdata.unit import econvs
from typing import List

from ase.geometry.cell import cell_to_cellpar
from ase import Atoms
import numpy as np

import ase.io
import dpdata
import re


from .util import LammpsData
from ai2_kit.core.log import get_logger

logger = get_logger(__name__)


def set_dplr_ext_from_cp2k_output(dp_sys: dpdata.LabeledSystem,
                                  cp2k_output: str,
                                  wannier_file: str,
                                  type_map: List[str],
                                  sys_charge_map: List[int],
                                  model_charge_map: List[int],
                                  ewald_h: float,
                                  ewald_beta: float,
                                  ext_efield,
                                  sel_type: List[int],
                                  ):

    wannier_atoms = ase.io.read(wannier_file)
    with open(cp2k_output, 'r') as fp:
        raw_energy = get_cp2k_output_total_energy(fp)

    ext_efield = np.reshape(ext_efield, [1, 3])
    natoms = dp_sys.get_natoms()
    nframes = dp_sys.get_nframes()
    assert nframes == 1, 'Only support one frame'

    symbols = np.array(dp_sys.data["atom_names"])[dp_sys.data["atom_types"]]
    sel_ids = get_sel_ids(dp_sys, type_map, sel_type)

    atomic_dipole, extended_coords = get_atomic_dipole(dp_sys, sel_ids, wannier_atoms)
    # get extended charges
    extended_charges = np.zeros(len(extended_coords))
    for symbol, ion_charge in zip(type_map, sys_charge_map):
        extended_charges[np.where(symbols == symbol)[0]] = ion_charge
    sel_symbols = symbols[sel_ids]
    for ii, wannier_charge in zip(sel_type, model_charge_map):
        sel_symbol = type_map[ii]
        extended_charges[np.where(sel_symbols == sel_symbol)[0] + natoms] = wannier_charge
    total_dipole = np.sum(extended_coords * extended_charges.reshape(-1, 1), axis=0)
    energy = raw_energy - np.inner(total_dipole, ext_efield.reshape(3))
    dp_sys.data['energies'] = np.reshape(energy, [-1])
    pbc_efield = get_pbc_atomic_efield(dp_sys, extended_coords, extended_charges, ewald_h, ewald_beta)
    # natoms * 3
    _efield = pbc_efield + ext_efield
    aparam = np.linalg.norm(_efield, axis=-1)
    efield = np.array(_efield) / aparam.reshape(natoms, 1)

    dp_sys.data['ext_efield'] = ext_efield.reshape(nframes, 3)
    dp_sys.data['efield'] = efield.reshape(nframes, natoms, 3)
    dp_sys.data['aparam'] = aparam.reshape(nframes, natoms, -1)
    dp_sys.data['atomic_dipole'] = atomic_dipole.reshape(nframes, -1)
    return dp_sys

def get_cp2k_output_total_energy(fp):
    start_pattern = r"  Total charge density g-space grids:"
    end_pattern = r"  Total energy:"
    start_pattern = re.compile(start_pattern)
    end_pattern = re.compile(end_pattern)

    flag = False
    data_lines = []
    nloop = 0
    for line in fp:
        line = line.strip('\n')
        if start_pattern.match(line):
            flag = True
        elif flag and end_pattern.match(line):
            flag = False
            nloop += 1

        if flag is True:
            data_lines.append(line)
    data_lines = np.reshape(data_lines, (nloop, -1))

    tot_e = 0.
    # grep the data from the last iteration
    for kw in data_lines[-1, 2:-1]:
        kw = kw.split(":")
        k = kw[0].strip(' ')
        v = float(kw[1]) * econvs["hartree"]
        if k != "Fermi energy":
            tot_e += v
    return tot_e


def get_sel_ids(dp_sys, type_map, sel_type):
    symbols = np.array(dp_sys.data["atom_names"])[dp_sys.data["atom_types"]]
    sel_ids = [np.where(symbols == type_map[ii])[0] for ii in sel_type]
    return np.concatenate(sel_ids)


def get_atomic_dipole(dp_sys, sel_ids, wannier_atoms, wannier_cutoff = 1.0):
    from MDAnalysis.lib.distances import distance_array, minimize_vectors

    coords = dp_sys.data['coords'].reshape(-1, 3)
    cellpar = cell_to_cellpar(dp_sys.data['cells'].reshape(3, 3))
    extended_coords = coords.copy()

    ref_coords = coords[sel_ids].reshape(-1, 3)
    e_coords = wannier_atoms.get_positions()
    dist_mat = distance_array(ref_coords, e_coords, box=cellpar)
    atomic_dipole = []
    for ii, dist_vec in enumerate(dist_mat):
        mask = (dist_vec < wannier_cutoff)
        cn = np.sum(mask)
        if cn != 4:
            raise ValueError(f'wannier atoms {ii} has {cn} atoms in the cutoff range')
        wc_coord_rel = e_coords[mask] - ref_coords[ii]
        wc_coord_rel = minimize_vectors(wc_coord_rel, box=cellpar)
        _atomic_dipole = wc_coord_rel.mean(axis=0)
        atomic_dipole.append(_atomic_dipole)
        wc_coord = _atomic_dipole + ref_coords[ii]
        extended_coords = np.concatenate((extended_coords, wc_coord.reshape(1, 3)), axis=0)
    atomic_dipole = np.reshape(atomic_dipole, (-1, 3))
    assert atomic_dipole.shape[0] == len(sel_ids)
    return atomic_dipole, extended_coords


def get_pbc_atomic_efield(dp_sys,
                          extended_coords,
                          extended_charges,
                          ewald_h,
                          ewald_beta,
                          ):
    from deepmd.infer.ewald_recp import EwaldRecp

    er = EwaldRecp(ewald_h, ewald_beta)
    natoms = dp_sys.get_natoms()
    cell = dp_sys.data['cells']
    e, f, v = er.eval(extended_coords.reshape(1, -1),
                        extended_charges.reshape(1, -1),
                        cell.reshape(1, -1))
    extended_efield = f.reshape(-1, 3) / extended_charges.reshape(-1, 1)
    # natoms * 3
    pbc_efield = extended_efield[:natoms]
    return pbc_efield


def get_sel_type(model_path) -> List[int]:
    try:
        from deepmd.infer import DeepDipole
    except ImportError:
        raise ImportError('deepmd-kit is not installed')

    dp = DeepDipole(model_path)
    return [t for t in dp.tselt]


def build_sel_type_assertion(sel_type, model_path: str, py_cmd='python'):
    return f'''{py_cmd} -c "from deepmd.infer import DeepDipole;dp = DeepDipole({repr(model_path)});assert{repr(sel_type)}==[t for t in dp.tselt]"'''


def dump_dplr_lammps_data(fp, atoms: Atoms, type_map: List[str], sel_type: List[int],
                          sys_charge_map: List[float], model_charge_map: List[float]):

    """
    dump atoms to LAMMPS data file for DPLR
    the naming convention of params follows Deepmd-Kit's

    about dplr: https://docs.deepmodeling.com/projects/deepmd/en/master/model/dplr.html
    about fitting tensor: https://docs.deepmodeling.com/projects/deepmd/en/master/model/train-fitting-tensor.html

    :param fp: file pointer
    :param type_map: the type map of atom type, for example, [O,H]
    :param sel_type: the selected type of atom, for example, [0] means atom type 0, aka O is selected
    :param sys_charge_map: the charge map of atom in system, for example, [6, 1]
    :param model_charge_map: the charge map of atom in model, for example, [-8]
    """

    if len(type_map) != len(sys_charge_map):
        raise ValueError(f'type_map {type_map} and sys_charge_map {sys_charge_map} must have the same length')

    if len(sel_type) != len(model_charge_map):
        raise ValueError(f'sel_type {sel_type} and model_charge_map {model_charge_map} must have the same length')

    new_atoms = atoms.copy()
    vtypes = get_unused_symbols(type_map, len(sel_type))

    r_atom_ids = []
    v_atom_ids = []

    # add virtual atoms
    for atype, vtype in zip(sel_type, vtypes):
        # create virtual atoms from the original atom type
        v_atoms = atoms[atoms.symbols == type_map[atype]]
        v_atoms.set_chemical_symbols([vtype] * len(v_atoms))  # type: ignore

        # create bonds between atom and its virtual atoms
        r_atom_ids.extend(np.where(atoms.symbols == type_map[atype])[0])
        v_atom_ids.extend(np.arange(len(v_atoms)) + len(new_atoms))  # type: ignore
        # this should be last
        new_atoms += v_atoms

    # build charges
    charges = np.zeros(len(new_atoms))
    for atype, charge in zip(type_map + vtypes, sys_charge_map + model_charge_map):
        charges[new_atoms.symbols==atype] = charge
    # build bonds
    n_bonds = len(r_atom_ids)
    # note: the index of lammps start from 1
    bonds = np.array([
        np.arange(n_bonds) + 1,  # bond id
        [1] * n_bonds,  # bond type
        np.array(r_atom_ids) + 1,  # bond left
        np.array(v_atom_ids) + 1,  # bond right
    ], dtype=int).T

    # write lammps data
    lmp_data = LammpsData(new_atoms)
    lmp_data.charges = charges
    lmp_data.bonds = bonds
    lmp_data.write(fp, atom_style='full', specorder=type_map + vtypes)

def get_unused_symbols(used_symbols: List[str], size: int):
    """
    Get unused symbols from ase

    :param used_symbols: symbols that are already used
    :param size: number of unused symbols to get
    """
    from ase.data import chemical_symbols

    unused_symbols = []
    for symbol in reversed(chemical_symbols):
        if symbol not in used_symbols:
            unused_symbols.append(symbol)
        if len(unused_symbols) == size:
            break
    return unused_symbols



from ai2_kit.core.util import ensure_dir, expand_globs, slice_from_str, SAMPLE_METHOD, list_sample
from ai2_kit.domain.cp2k import dump_coord_n_cell

from typing import List, Union, Optional
from ase import Atoms
import ase.io


class AseTool:

    def __init__(self, atoms_list: Optional[List[Atoms]] = None):
        self._atoms_list: List[Atoms] = [] if atoms_list is None else atoms_list

    def read(self, *file_path_or_glob: str, **kwargs):
        files = expand_globs(file_path_or_glob)
        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._read(file, **kwargs)
        return self

    def write(self, filename: str, **kwargs):
        ensure_dir(filename.format(i=0))
        self._write(filename, self._atoms_list, **kwargs)

    def set_cell(self, cell, scale_atoms=False, apply_constraint=True):
        for atoms in self._atoms_list:
            atoms.set_cell(cell, scale_atoms=scale_atoms, apply_constraint=apply_constraint)
        return self

    def set_pbc(self, pbc):
        for atoms in self._atoms_list:
            atoms.set_pbc(pbc)
        return self

    def set_by_ref(self, ref_file: str, **kwargs):
        kwargs.setdefault('index', 0)
        ref_atoms = ase.io.read(ref_file, **kwargs)
        assert isinstance(ref_atoms, Atoms), 'Only support single frame reference'
        for atoms in self._atoms_list:
            try:
                atoms.set_cell(ref_atoms.get_cell())
            except ValueError:
                pass
            try:
                atoms.set_pbc(ref_atoms.get_pbc())
            except ValueError:
                pass
        return self

    def size(self):
        """
        size of loaded data
        """
        print(len(self._atoms_list))
        return self

    def slice(self, expr: str):
        """
        slice systems by python slice expression, for example
        `10:`, `:10`, `::2`, etc

        :param start: start index
        :param stop: stop index
        :param step: step
        """
        s = slice_from_str(expr)
        self._atoms_list = self._atoms_list[s]
        return self

    def sample(self, size: int, method: SAMPLE_METHOD='even', **kwargs):
        """
        sample data

        :param size: size of sample, if size is larger than data size, return all data
        :param method: method to sample, can be 'even', 'random', 'truncate', default is 'even'
        :param seed: seed for random sample, only used when method is 'random'
        """
        self._atoms_list= list_sample(self._atoms_list, size, method, **kwargs)
        return self

    def to_dpdata(self, labeled=False):
        """
        convert to dpdata format and use dpdata tool to handle

        :param labeled: if True, use dpdata.LabeledSystem, else use dpdata.System
        """
        from .dpdata import dpdata, DpdataTool
        System = dpdata.LabeledSystem if labeled else dpdata.System
        systems = [System(atom, fmt='ase/structure') for atom in self._atoms_list]
        return DpdataTool(systems=systems)

    def delete_atoms(self, id: Union[int, List[int]], start_id=0):
        """
        delete atoms by id or list of id

        :param id: id or list of id
        :param start_id: the start id of first item, for example, in LAMMPS the id of first item is 1, in ASE it is 0
        """

        ids = [id] if isinstance(id, int) else id
        for atoms in self._atoms_list:
            for i in sorted(ids, reverse=True):
                assert i >= start_id, f'Invalid id {i}'
                del atoms[i - start_id]
        return self

    def write_each_frame(self, filename: str, **kwargs):
        """
        write each frame to a separate file

        :param filename: the filename template, use {i} to represent the index, for example, 'frame_{i}.xyz'
        :param kwargs: other arguments for ase.io.write
        """
        ensure_dir(filename.format(i=0))
        for i, atoms in enumerate(self._atoms_list):
            self._write(filename.format(i=i), atoms, **kwargs)


    def write_dplr_lammps_data(self, filename: str,
                               type_map: List[str], sel_type: List[int],
                               sys_charge_map: List[float], model_charge_map: List[float]):

        """
        write atoms to LAMMPS data file for DPLR
        the naming convention of params follows Deepmd-Kit's

        about dplr: https://docs.deepmodeling.com/projects/deepmd/en/master/model/dplr.html
        about fitting tensor: https://docs.deepmodeling.com/projects/deepmd/en/master/model/train-fitting-tensor.html


        :param filename: the filename of LAMMPS data file, use {i} to represent the index, for example, 'frame_{i}.lammps.data'
        :param type_map: the type map of atom type, for example, [O,H]
        :param sel_type: the selected type of atom, for example, [0] means atom type 0, aka O is selected
        :param sys_charge_map: the charge map of atom in system, for example, [6, 1]
        :param model_charge_map: the charge map of atom in model, for example, [-8]
        """

        from ai2_kit.domain.dpff import dump_dplr_lammps_data
        ensure_dir(filename.format(i=0))
        for i, atoms in enumerate(self._atoms_list):
            with open(filename.format(i=i), 'w') as f:
                dump_dplr_lammps_data(f, atoms=atoms, type_map=type_map, sel_type=sel_type,
                                      sys_charge_map=sys_charge_map, model_charge_map=model_charge_map)


    def _read(self, filename: str, **kwargs):
        kwargs.setdefault('index', ':')
        data = ase.io.read(filename, **kwargs)
        if not isinstance(data, list):
            data = [data]
        self._atoms_list += data

    def _write(self, filename: str, atoms_list, **kwargs):
        assert len(atoms_list) > 0, 'No atoms to write'
        fmt = kwargs.get('format', None)
        if fmt == 'lammps-dump-text':
            return self._write_lammps_dump_text(filename, atoms_list, **kwargs)
        if fmt == 'cp2k-inc' or (fmt is None and filename.endswith('.inc')):
            assert len(atoms_list) == 1, 'cp2k-inc only support single frame'
            return self._write_cp2k_inc(filename, atoms_list[0])
        return ase.io.write(filename, atoms_list, **kwargs)

    def _write_cp2k_inc(self, filename: str, atoms: Atoms):
        with open(filename, 'w') as f:
            dump_coord_n_cell(f, atoms)

    def _write_lammps_dump_text(self, filename: str, atoms_list, **kwargs):
        if not isinstance(atoms_list, list):
            atoms_list = [atoms_list]
        type_map = kwargs.get('type_map')
        lines = []
        for i, atoms in enumerate(atoms_list):
            lines += [
                "ITEM: TIMESTEP", str(i),
                "ITEM: NUMBER OF ATOMS", str(len(atoms)),
            ]
            # It is kind of hacky to set boundary and cell in this way.
            # TODO: refactor it in the future.
            pbc = atoms.get_pbc()
            boundary =  ' '.join(['pp' if b else 'ff' for b in pbc])
            cell = atoms.get_cell()
            lines += [
                f"ITEM: BOX BOUNDS xy xz yz {boundary}",
                "%.16e %.16e %.16e" % (0, cell[0][0], 0),
                "%.16e %.16e %.16e" % (0, cell[1][1], 0),
                "%.16e %.16e %.16e" % (0, cell[2][2], 0),
                # TODO: support more properties, forces for example
                "ITEM: ATOMS id type x y z",
            ]

            for j, atom in enumerate(atoms):
                pos = atom.position
                no = atom.number if type_map is None else type_map.index(atom.symbol) + 1
                lines.append("%3d %2s %10.5f %10.5f %10.5f" % (j+1, no , pos[0], pos[1], pos[2]))

        with open(filename, 'w') as f:
            f.write('\n'.join(lines))



from ai2_kit.feat.spectrum.viber import dpdata_read_cp2k_viber_data
from ai2_kit.core.util import ensure_dir, expand_globs, list_sample, SAMPLE_METHOD, perf_log, slice_from_str
from ai2_kit.core.log import get_logger

from typing import Optional
from dpdata.data_type import Axis, DataType
import numpy as np
import dpdata


logger = get_logger(__name__)


def register_data_types():
    if getattr(dpdata, '__registed__', False):
        return

    DATA_TYPES = [
        DataType("fparam", np.ndarray, (Axis.NFRAMES, -1), required=False),  # type: ignore
        DataType("aparam", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, -1), required=False), # type: ignore
        DataType("efield", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, 3), required=False), # type: ignore
        DataType("ext_efield", np.ndarray, (Axis.NFRAMES, 3), required=False), # type: ignore
        DataType("atomic_dipole", np.ndarray, (Axis.NFRAMES, -1), required=False), # type: ignore
        DataType("atomic_polarizability", np.ndarray, (Axis.NFRAMES, -1), required=False), # type: ignore
    ]
    dpdata.System.register_data_type(*DATA_TYPES) # type: ignore
    dpdata.LabeledSystem.register_data_type(*DATA_TYPES) # type: ignore
    dpdata.__registed__ = True  # type: ignore


register_data_types()


class DpdataTool:

    def __init__(self, verbose = False, systems: Optional[list] = None):
        self._systems = [] if systems is None else systems
        self._verbose = verbose

    def read(self, *file_path_or_glob: str, **kwargs):
        """
        read data from multiple paths, support glob pattern
        default format is deepmd/npy

        :param file_path_or_glob: path or glob pattern to find data files
        :param kwargs: arguments to pass to dpdata.System / dpdata.LabeledSystem
        """
        kwargs.setdefault('fmt', 'deepmd/npy')
        files = expand_globs(file_path_or_glob)
        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._read(file, **kwargs)
        return self

    def filter(self, lambda_expr: str):
        """
        filter data with lambda expression

        :param lambda_expr: lambda expression to filter data
        """
        fn = eval(lambda_expr)
        self._systems = [system for system in self._systems if fn(system.data)]
        return self

    def slice(self, expr: str):
        """
        slice systems by python slice expression, for example
        `10:`, `:10`, `::2`, etc

        :param start: start index
        :param stop: stop index
        :param step: step
        """
        s = slice_from_str(expr)
        self._systems = self._systems[s]
        return self

    def sample(self, size: int, method: SAMPLE_METHOD='even', **kwargs):
        """
        sample data

        :param size: size of sample, if size is larger than data size, return all data
        :param method: method to sample, can be 'even', 'random', 'truncate', default is 'even'
        :param seed: seed for random sample, only used when method is 'random'
        """
        self._systems = list_sample(self._systems, size, method, **kwargs)
        return self

    def size(self):
        """
        size of loaded data
        """
        print(len(self._systems))
        return self

    def write(self, out_path: str, fmt='deepmd/npy', merge: bool = True):
        """
        write data to specific path, support deepmd/npy, deepmd/raw, deepmd/hdf5 formats
        :param out_path: path to write data
        :param fmt: format to write, default is deepmd/npy
        :param merge: if True, merge all data use dpdata.MultiSystems, else write data without merging
        """
        ensure_dir(out_path)
        if len(self._systems) == 0:
            raise ValueError('No data to merge')
        if merge:
            systems = dpdata.MultiSystems(self._systems[0])
        else:
            systems = self._systems[0]

        for system in self._systems[1:]:
            systems.append(system)

        if fmt == 'deepmd/npy':
            systems.to_deepmd_npy(out_path)  # type: ignore
        elif fmt == 'deepmd/raw':
            systems.to_deepmd_raw(out_path)  # type: ignore
        elif fmt == 'deepmd/hdf5':
            systems.to_deepmd_hdf5(out_path)  # type: ignore
        else:
            raise ValueError(f'Unknown fmt {fmt}')

    def set_fparam(self, fparam):
        """
        Set fparam for all systems

        :param fparam: fparam to set, should be a scalar or vector, e.g. 1.0 or [1.0, 2.0]
        """
        for system in self._systems:
            set_fparam(system, fparam)
        return self

    def eval(self, dp_model: str):
        """
        Use deepmd model to label energy, force and viral

        :param dp_model: path to deepmd frozen model
        """
        from deepmd.infer import DeepPot
        systems = dpdata.System()
        systems.extend(self._systems)

        coords = systems.data['coords']
        cells = None if systems.nopbc else systems.data['cells']
        atypes = systems.data['atom_types']

        perf_log('before deepmd label')
        pot = DeepPot(dp_model, auto_batch_size=True)  # type: ignore
        e, f, v = pot.eval(coords=coords, cells=cells, atom_types=atypes)  # type: ignore
        perf_log('after deepmd label')

        n_atoms = systems.get_natoms()
        n_frames = systems.get_nframes()

        e = e.reshape((n_frames,))
        f = f.reshape((n_frames, n_atoms, 3))
        v = v.reshape((n_frames, 3, 3))

        data = {**systems.data, 'energies': e, 'forces': f, "virials": v}
        # replace system files
        self._systems = []
        self._systems.extend(dpdata.LabeledSystem.from_dict({'data':data}))  # type: ignore
        perf_log('after update data')
        return self

    def to_ase(self):
        """
        Convert dpdata format to ase format, and use ase tool to handle
        """
        from .ase import AseTool
        atoms_list = []
        for sys in self._systems:
            atoms_list.extend(sys.to_ase_structure())
        return AseTool(atoms_list=atoms_list)

    def _read(self, data_path: str, **kwargs):
        fmt = kwargs.pop('fmt')
        label = kwargs.get('label', True)

        assert fmt is not None, 'fmt is required'
        if fmt == 'cp2k/viber':
            try:
                system = dpdata_read_cp2k_viber_data(data_path, **kwargs)
            except:
                logger.warn(f'Fail to read {data_path}')
                return
        else:
            system = dpdata.LabeledSystem(data_path, fmt=fmt, **kwargs) if label else dpdata.System(data_path, fmt=fmt, **kwargs)

        fparam = kwargs.get('fparam', None)
        if fparam is not None:
            set_fparam(system, fparam)

        # use extend to flatten system
        # TODO: I don't know if flatten system will lead to performance issues.
        self._systems.extend(system)  # type: ignore

    def _verbose_log(self, msg, **kwargs):
        if self._verbose:
            logger.info(msg, **kwargs)


def set_fparam(system, fparam):
    nframes = system.get_nframes()
    system.data['fparam'] = np.tile(fparam, (nframes, 1))
    return system


from MDAnalysis.lib.distances import distance_array
import MDAnalysis as mda
import numpy as np
import fire
import os


def _ecn_analysis_per_frame(frame_index, reference_group, configuration_group, universe):
    l_avg, ecn = 0, 0
    universe.trajectory[frame_index]
    distance_mat = distance_array(reference=reference_group, configuration=configuration_group, box=universe.dimensions)
    # Calculate l_av (average bond length).
    n_in, n_o = distance_mat.shape
    l_min_array = np.min(distance_mat, axis=1)
    l_av = np.divide(distance_mat,l_min_array[:, np.newaxis])
    l_av = np.exp(1.0 - np.power(l_av, 6))
    denominator = np.sum(l_av, axis=1)
    l_av = np.divide(np.multiply(distance_mat, l_av), denominator[:, np.newaxis])
    l_av = np.sum(l_av, axis=1)
    l_av = np.mean(l_av)
    # Calculate ECN (effective coordination number)
    ecn_part = l_av.copy()
    ecn_part = np.exp(1.0- np.power(np.divide(distance_mat, ecn_part), 6))
    ecn_part = np.sum(ecn_part)/n_in
    l_avg += l_av
    ecn += ecn_part
    return frame_index, l_avg, ecn


def ecn_analysis(input_traj: str, out_dir: str, center: str, ligand: str, cell: list[float]):
    """
    Calculate ECN (effective coordination number) and l_av (average bond length).

    :param input_traj: path of input trajectory file
    :param out_dir: path of output directory
    :param center: central atoms, in the format of MDA atom selection string, e.g. 'name In'
    :param ligand: atoms that surround the central atoms, in the format of MDA atom selection string, e.g. 'name O'
    :param cell: cell dimensions
    """

    universe = mda.Universe(input_traj)
    universe.dimensions = np.array(cell)
    atomgroup = universe.atoms
    reference_group = atomgroup.select_atoms(center)
    configuration_group = atomgroup.select_atoms(ligand)
    # TODO: this can be parallelized
    results = []
    for k in range(universe.trajectory.n_frames):
        result = _ecn_analysis_per_frame(frame_index=k,
                                         reference_group=reference_group,
                                         configuration_group=configuration_group,
                                         universe=universe)
        results.append(result)
    results_arr = np.array(results)
    avr_arr = np.mean(results_arr, axis=0)

    # dump results
    os.makedirs(out_dir, exist_ok=True)
    np.savetxt(os.path.join(out_dir, 'raw.txt'), results_arr,
               fmt='    %d        %.4f   %.4f',
               delimiter='   ', header='frame_index   l_av   ECN')
    with open(os.path.join(out_dir, 'stats.txt'), mode='w') as fp:
        fp.write('\n'.join([f'l_av = {avr_arr[1]:.4f}',
           f'ECN = {avr_arr[2]:.4f}']))


def _count_shared_polyhedra_per_frame(frame_index, reference_group, configuration_group, universe, cutoff, coordination_num = -1):
    universe.trajectory[frame_index]

    distance_mat = distance_array(reference=reference_group, configuration=configuration_group, box=universe.dimensions)
    neighbor_mat = distance_mat < cutoff
    neighbor_mat = neighbor_mat.astype(int)
    # common_neighbor_mat: a matrix of coordination and shared oxygen numbers.
    # The former numbers are diagonal and the later are other elements.
    common_neighbor_mat = np.dot(neighbor_mat, neighbor_mat.T)

    # Counting polyhedra with particular coordination number.
    if coordination_num > 0:
        for i in range(reference_group.n_atoms):
            if common_neighbor_mat[i][i] != coordination_num:
                common_neighbor_mat[i] = 0
                common_neighbor_mat[:, i] = 0
    # counting
    common_neighbor_mat = common_neighbor_mat[np.triu_indices_from(common_neighbor_mat, k=1)]
    corner = np.count_nonzero(common_neighbor_mat == 1)
    edge = np.count_nonzero(common_neighbor_mat == 2)
    face = np.count_nonzero(common_neighbor_mat>=3)
    return frame_index, corner, edge, face


def count_shared_polyhedra(input_traj: str, out_dir: str, center: str, ligand: str, cell: list[float], cutoff: float, coord_num: int = -1):
    """
    Count the number of shared polyhedra in each frame of the trajectory.

    :param input_traj: path of input trajectory file
    :param out_dir: path of output directory
    :param center: central atoms, in the format of MDA atom selection string, e.g. 'name In'
    :param ligand: atoms that surround the central atoms, in the format of MDA atom selection string, e.g. 'name O'
    :param cell: cell dimensions
    :param cutoff: cutoff distance
    :param coord_num: coordination number
    """
    universe = mda.Universe(input_traj)
    universe.dimensions = np.array(cell)
    atomgroup = universe.atoms
    reference_group = atomgroup.select_atoms(center)
    configuration_group = atomgroup.select_atoms(ligand)

    # TODO: this can be parallelized
    results = []
    for k in range(universe.trajectory.n_frames):
        result = _count_shared_polyhedra_per_frame(frame_index=k,
                                                   reference_group=reference_group,
                                                   configuration_group=configuration_group,
                                                   universe=universe,
                                                   cutoff=cutoff, coordination_num=coord_num)
        results.append(result)

    results_arr = np.array(results)
    avg_arr = np.mean(results_arr, axis=0)

    # dump results
    os.makedirs(out_dir, exist_ok=True)
    np.savetxt(os.path.join(out_dir, 'raw.txt'), results_arr,
               fmt='    %d        %.4f   %.4f   %.4f',
               delimiter='   ', header='frame_index   corner   edge   face')

    with open(os.path.join(out_dir, 'stats.txt'), mode='w') as fp:
        fp.write('\n'.join([f'Corner-share = {avg_arr[1]:.4f}',
                            f'Edge-share = {avg_arr[2]:.4f}',
                            f'Face-share = {avg_arr[3]:.4f}', ]))


cli_entry = {
    'ecn-analysis': ecn_analysis,
    'count-shared-polyhedra': count_shared_polyhedra,
}


if __name__ == '__main__':
    fire.Fire(cli_entry)



from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import dump_json, dump_text, flush_stdio, limit
from ai2_kit.core.pydantic import BaseModel

from typing import List, Optional, Tuple, Dict
from io import StringIO
from dataclasses import dataclass
import pandas as pd
from tabulate import tabulate
from itertools import groupby
from functools import lru_cache
import traceback

import ase.io
import os

from .data import get_data_format, DataFormat, artifacts_to_ase_atoms
from .iface import ICllSelectorOutput, BaseCllContext
from .constant import LAMMPS_DUMP_DIR, LAMMPS_DUMP_SUFFIX, DEFAULT_ASAP_SOAP_DESC, DEFAULT_ASAP_PCA_REDUCER


logger = get_logger(__name__)


class CllModelDeviSelectorInputConfig(BaseModel):

    class AsapOptions(BaseModel):
        disable: bool = False
        limit_per_cluster: int = 1
        """
        limit the number of structures to be selected from the same cluster
        """
        sort_by_ssw_energy: bool = False
        """
        sorted the structures by ssw_energy in each cluster
        """
        descriptor: dict = {'soap': { **DEFAULT_ASAP_SOAP_DESC, 'preset': 'minimal'}}
        dim_reducer: dict = {'pca': DEFAULT_ASAP_PCA_REDUCER}
        cluster: dict = {'dbscan': {}}

    f_trust_lo: float = 0.
    """
    the lower bound of model_devi score to select the structure for labeling
    """
    f_trust_hi: float = 65535.
    """
    the upper bound of model_devi score to select the structure for labeling
    """
    new_explore_system_q: float = 0.25
    """
    the quantile of model_devi score to select the structure for next round of exploration
    """
    asap_options: Optional[AsapOptions]
    """
    options for ASAP to further select candidates
    """
    screening_fn: Optional[str] = None
    """
    the function to screen the candidates, e.g
    "lambda x: x['ssw_energy'] < -1000"
    """
    max_decent_per_traj: int = -1
    """
    limit the max number of decent structures per trajectory, -1 means unlimited
    """
    workers: int = 4
    """
    number of workers to run the analysis
    """


@dataclass
class CllModelDevSelectorContext(BaseCllContext):
    ...


@dataclass
class CllModelDeviSelectorOutput(ICllSelectorOutput):
    candidates: List[Artifact]
    passing_rate: float
    new_explore_systems: List[Artifact]

    def get_model_devi_dataset(self):
        return self.candidates

    def get_passing_rate(self) -> float:
        return self.passing_rate

    def get_new_explore_systems(self) -> List[Artifact]:
        return self.new_explore_systems


@dataclass
class CllModelDeviSelectorInput:
    config: CllModelDeviSelectorInputConfig
    model_devi_data: List[Artifact]
    model_devi_file: str
    type_map: List[str]

    def set_model_devi_dataset(self, data: List[Artifact]):
        self.model_devi_data = data


async def cll_model_devi_selector(input: CllModelDeviSelectorInput, ctx: CllModelDevSelectorContext):
    executor = ctx.resource_manager.default_executor
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    executor.mkdir(work_dir)

    f_trust_lo = input.config.f_trust_lo
    f_trust_hi = input.config.f_trust_hi

    results = executor.run_python_fn(bulk_select_structures_by_model_devi)(
        model_devi_outputs=[a.to_dict() for a in input.model_devi_data],
        model_devi_file=input.model_devi_file,
        f_trust_lo=f_trust_lo, f_trust_hi=f_trust_hi,
        type_map=input.type_map, work_dir=work_dir,
        new_explore_system_q=input.config.new_explore_system_q,
        max_decent_per_traj=input.config.max_decent_per_traj,
        screening_fn=input.config.screening_fn,
        workers=input.config.workers,
    )

    candidates = [ result['decent'] for result, _ in results if 'decent' in result ]
    new_systems = [ result['next'] for result, _ in results if 'next' in result ]
    stats = [ stats for _, stats in results ]

    # group next_structures by `attrs.source` and keep the first one
    # so that the total number of explore structures will be the same as the original one
    get_source = lambda s: s['attrs']['source']
    # filter out the structures whose ancestor ends with '-fin'
    # FIXME: use a dedicated field to indicate the final structure
    new_systems = [s for s in new_systems if not s['attrs']['ancestor'].endswith('-fin')]
    new_systems = sorted(new_systems, key=get_source)
    new_systems = [next(group) for _source, group in groupby(
        new_systems, key=get_source)]

    # write model_deviation stats report
    # TODO: refactor into a function
    def _get_row(stat: dict):
        url = stat['src']
        url = f'...{url[-30:]}' if len(url) > 30 else url  # wrap url if it is too long
        total, good, decent, poor = stat['total'], stat['good'], stat['decent'], stat['poor']
        return [
            url, total, good, decent, poor,
            f'{good / total * 100:.2f}%', f'{decent / total * 100:.2f}%', f'{poor / total * 100:.2f}%',
        ]
    headers = ['file', 'total', 'good', 'decent', 'poor', 'good%', 'decent%', 'poor%']
    table = [ _get_row(stat) for stat in stats]
    total = sum(row[1] for row in table)
    total_good = sum(row[2] for row in table)

    stats_report = tabulate(table, headers=headers, tablefmt='tsv')
    logger.info('stats report: \n%s\n', stats_report)
    executor.dump_text(stats_report, os.path.join(work_dir, 'stats.tsv'))

    # further select candidates by ASAP
    if input.config.asap_options and not input.config.asap_options.disable:
        asap_options = input.config.asap_options
        candidates = executor.run_python_fn(bulk_select_distinct_structures)(
            candidates=candidates,
            descriptor_opt=asap_options.descriptor,
            dim_reducer_opt=asap_options.dim_reducer,
            cluster_opt=asap_options.cluster,
            type_map=input.type_map,
            work_dir=work_dir,
            limit_per_cluster=asap_options.limit_per_cluster,
            sort_by_energy=asap_options.sort_by_ssw_energy,
            workers=input.config.workers,
        )

    return CllModelDeviSelectorOutput(
        candidates=[Artifact.of(**a) for a in candidates],
        new_explore_systems=[Artifact.of(**a) for a in new_systems],
        passing_rate=total_good / total,
    )


def bulk_select_structures_by_model_devi(model_devi_outputs: List[ArtifactDict],
                                            model_devi_file: str,
                                            f_trust_lo: float,
                                            f_trust_hi: float,
                                            new_explore_system_q: float,
                                            type_map: List[str],
                                            work_dir: str,
                                            max_decent_per_traj: int,
                                            screening_fn: Optional[str],
                                            workers: int = 4,
                                            ) -> List[Tuple[Dict[str, ArtifactDict], dict]]:
    import joblib
    return joblib.Parallel(n_jobs=workers)(
        joblib.delayed(select_structures_by_model_devi)(
            model_devi_output=output,
            model_devi_file=model_devi_file,
            f_trust_lo=f_trust_lo, f_trust_hi=f_trust_hi,
            type_map=type_map,
            work_dir=os.path.join(work_dir, 'model_devi', f'{i:06}'),
            max_decent_per_traj=max_decent_per_traj,
            new_explore_system_q=new_explore_system_q,
            screening_fn=screening_fn,
        )
        for i, output in enumerate(model_devi_outputs)
    )  # type: ignore


def select_structures_by_model_devi(model_devi_output: ArtifactDict,
                                    model_devi_file: str,
                                    f_trust_lo: float,
                                    f_trust_hi: float,
                                    type_map: List[str],
                                    work_dir: str,
                                    new_explore_system_q: float,
                                    max_decent_per_traj: int,
                                    screening_fn: Optional[str],
                                    ) -> Tuple[Dict[str, ArtifactDict], dict]:
    """
    analysis the model_devi output of explore stage and select candidates

    :param next_explore_system_q: the quantile of model_devi score to select the structure for next round of exploration
    """
    os.makedirs(work_dir, exist_ok=True)
    dump_json(model_devi_output, os.path.join(work_dir, 'model_devi_output.debug.json'))

    model_devi_dir = model_devi_output['url']
    model_devi_file = model_devi_output['attrs'].pop('model_devi_file', model_devi_file)

    force_col = 'max_devi_f'
    print(f'criteria: {f_trust_lo} <= {force_col} < {f_trust_hi}')

    # get path of model_devi file
    data_format = get_data_format(model_devi_output)  # type: ignore
    if data_format in (DataFormat.LAMMPS_OUTPUT_DIR, DataFormat.LASP_LAMMPS_OUT_DIR):
        model_devi_file = os.path.join(model_devi_dir, model_devi_file)
    elif data_format == DataFormat.ANYWARE_OUTPUT_DIR:
        model_devi_file = os.path.join(model_devi_dir, 'model_devi.out')
    else:
        raise ValueError('unknown model_devi_data types')
    logger.info('start to analysis file: %s', model_devi_file)

    # load model_devi data
    with open(model_devi_file, 'r') as f:
        text = f.read()
    df = pd.read_csv(StringIO(text.lstrip('#')), delim_whitespace=True)
    # layout:
    #        step  max_devi_v  min_devi_v  avg_devi_v  max_devi_f  min_devi_f  avg_devi_f
    # 0        0    0.006793    0.000672    0.003490    0.143317    0.005612    0.026106
    # 1      100    0.006987    0.000550    0.003952    0.128178    0.006042    0.022608

    # load structures
    atoms_list = []
    if data_format == DataFormat.LAMMPS_OUTPUT_DIR:
        lammps_dump_dir = model_devi_output['attrs'].pop('lammps_dump_dir', LAMMPS_DUMP_DIR)
        for frame_id in df.step:
            dump_file = os.path.join(model_devi_dir, lammps_dump_dir, f'{frame_id}{LAMMPS_DUMP_SUFFIX}')
            atoms_list += ase.io.read(dump_file, ':', format='lammps-dump-text', specorder=type_map)
    elif data_format in (DataFormat.LASP_LAMMPS_OUT_DIR, DataFormat.ANYWARE_OUTPUT_DIR):
        structures_file = os.path.join(model_devi_dir, 'structures.xyz')
        atoms_list += ase.io.read(structures_file, ':', format='extxyz')
    else:
        raise ValueError('unknown model_devi_data types')

    # screening structure before model_devi analysis
    if screening_fn is not None:
        if 'ssw_energy' in atoms_list[0].info:
            s_ssw_energy = pd.Series(map(lambda atoms: atoms.info['ssw_energy'], atoms_list))  # type: ignore

            # the following ssw_* methods are for the screening_fn
            # so don't need to worry about the unused warning
            ssw_energy_max = s_ssw_energy.max()
            ssw_energy_min = s_ssw_energy.min()
            def ssw_energy_quantile(q):
                # the quantile will be evaluated every time, which is not efficient
                # so here we cache the result, carefully
                return lru_cache()(s_ssw_energy.quantile)(q)
            ssw_enenrgy_quantile = ssw_energy_quantile

        # return the df row whose atoms pass the screening_fn
        _screening_fn = eval(screening_fn, locals())  # str to function
        df = df[[ _screening_fn(atoms) for atoms in atoms_list ]]


    # evaluate new found structures by their model_devi score in 3 levels: good, decent, poor
    good_df   = df[df[force_col] < f_trust_lo]
    decent_df = df[(df[force_col] >= f_trust_lo) & (df[force_col] < f_trust_hi)]
    poor_df   = df[df[force_col] >= f_trust_hi]

    # select the last frame from df whose model_devi score is less than the quantile
    # as the initial structure for next round of exploration to replace the original one
    # the next frame should be selected from good or decent frame
    # if there is no good or decent frame, use the first frame as the initial structure
    # NOTE: equal is essential to ensure the existence of next structure
    # NOTE: select the last frame can increase the diversity of structures
    _ndf = df[df[force_col] < f_trust_hi]
    if len(_ndf) == 0:
        next_df = df.head(1)  # the first frame is the initial structure
    else:
        next_df = _ndf[_ndf[force_col] <= _ndf[force_col].quantile(new_explore_system_q)].tail(1)

    stats = {
        'src': model_devi_file,
        'total': len(df),
        'good': len(good_df),
        'decent': len(decent_df),
        'poor': len(poor_df),
    }

    result: Dict[str, ArtifactDict] = {}
    # TODO: refactor the following repeating code
    # TODO: dump good may lead to storage issue, so disable it for now
    if len(good_df) > 0:
        good_file = os.path.join(work_dir, 'good.xyz')
        # ase.io.write(good_file, [atoms_list[_i] for _i in good_df.index], format='extxyz')
        result['good'] = {'url': good_file, 'format': DataFormat.EXTXYZ,  # type: ignore
                            'attrs': {**model_devi_output['attrs']}}

    if len(poor_df) > 0:
        poor_file = os.path.join(work_dir, 'poor.xyz')
        # ase.io.write(poor_file, [atoms_list[_i] for _i in poor_df.index], format='extxyz')
        result['poor'] = {'url': poor_file, 'format': DataFormat.EXTXYZ,  # type: ignore
                            'attrs': {**model_devi_output['attrs']}}
    if len(decent_df) > 0:
        decent_file = os.path.join(work_dir, 'decent.xyz')
        ase.io.write(decent_file,
                        list(limit((atoms_list[_i] for _i in decent_df.index), max_decent_per_traj)),
                        format='extxyz')
        result['decent'] = {'url': decent_file, 'format': DataFormat.EXTXYZ,  # type: ignore
                            'attrs': {**model_devi_output['attrs']}}
    if len(next_df) > 0:
        next_file = os.path.join(work_dir, 'next.xyz')
        ase.io.write(next_file, [atoms_list[_i] for _i in next_df.index], format='extxyz')
        result['next'] = {'url': next_file, 'format': DataFormat.EXTXYZ,  # type: ignore
                            'attrs': {**model_devi_output['attrs']}}
    dump_json([result, stats, list(decent_df.index), list(next_df.index)], os.path.join(work_dir, 'result.debug.json'))
    return result, stats


def bulk_select_distinct_structures(candidates: List[ArtifactDict],
                                    descriptor_opt: dict,
                                    dim_reducer_opt: dict,
                                    cluster_opt: dict,
                                    type_map: List[str],
                                    work_dir: str,
                                    limit_per_cluster: int = -1,
                                    sort_by_energy: bool = False,
                                    workers: int = 4,
                                    ) -> List[ArtifactDict]:
    try:
        dump_json(candidates, os.path.join(work_dir, 'candidates.debug.json'))
    except Exception as e:
        pass
    get_ancestor = lambda c: c['attrs']['ancestor']
    candidates = sorted(candidates, key=get_ancestor)
    inputs = []
    for i, (ancestor_key, candidate_group) in enumerate(groupby(candidates, key=get_ancestor)):
        candidate_group = list(candidate_group)
        inputs.append((candidate_group, candidate_group[0]['attrs']))

    import joblib
    return joblib.Parallel(n_jobs=workers)(
        joblib.delayed(select_distinct_structures)(
            candidates=group,
            attrs=attrs,
            descriptor_opt=descriptor_opt,
            dim_reducer_opt=dim_reducer_opt,
            cluster_opt=cluster_opt,
            type_map=type_map,
            work_dir=os.path.join(work_dir, 'asap', f'{i:06}'),
            limit_per_cluster=limit_per_cluster,
            sort_by_energy=sort_by_energy,
        ) for i, (group, attrs) in enumerate(inputs)
    )  # type: ignore


def select_distinct_structures(candidates: List[ArtifactDict],
                                attrs: dict,
                                descriptor_opt: dict,
                                dim_reducer_opt: dict,
                                cluster_opt: dict,
                                type_map: List[str],
                                work_dir: str,
                                limit_per_cluster: int = -1,
                                sort_by_energy: bool = False,
                                ):

    from .asap import get_descriptor, reduce_dimension, get_trainer, get_cluster
    from asaplib.data.xyz import ASAPXYZ

    os.makedirs(work_dir, exist_ok=True)

    dump_json(attrs, os.path.join(work_dir, 'attrs.debug.json'))
    # load structures and save it to a tmp file for ASAP to load
    atoms_list = [atoms for _, atoms in artifacts_to_ase_atoms(candidates, type_map=type_map)]


    if len(atoms_list) < 20:
        # FIXME: there are a lot of potential issue when the number of atoms is small
        # the root cause is in asaplib, which I guess has not been tested with small dataset
        selected_atoms_list = atoms_list
    elif limit_per_cluster <= 0:
        selected_atoms_list = atoms_list
    else:
        if sort_by_energy and 'ssw_energy' in atoms_list[0].info:
            atoms_list = sorted(atoms_list, key=lambda atoms: atoms.info['ssw_energy'])

        # use asaplib to group structures
        # load structures to ASAP
        tmp_structures_file = os.path.join(work_dir, '.tmp-structures.xyz')
        ase.io.write(tmp_structures_file, atoms_list, format='extxyz')
        asapxyz = ASAPXYZ(tmp_structures_file)

        # group structures
        try:
            asap_path_prefix = os.path.join(work_dir, 'asap')
            descriptors, _ = get_descriptor(asapxyz, descriptor_opt, path_prefix=asap_path_prefix)
            reduced_descriptors = reduce_dimension(descriptors, dim_reducer_opt)
            trainer = get_trainer(reduced_descriptors, cluster_opt)
            cluster_labels = get_cluster(asapxyz, reduced_descriptors, trainer, path_prefix=asap_path_prefix)

            # dump_json(cluster_labels, os.path.join(work_dir, 'cluster.debug.json'))
            selected_frames = []
            for frames in cluster_labels.values():
                if len(frames) < limit_per_cluster:
                    selected_frames += list(frames)
                else:
                    selected_frames += list(frames[:limit_per_cluster])
            selected_atoms_list = [atoms_list[i] for i in selected_frames]
        except Exception as e:
            print('asaplib failed: %s', e)
            # dump exception to file
            dump_text(traceback.format_exc(), os.path.join(work_dir, 'asaplib-exception.txt'))
            selected_atoms_list = atoms_list

    # write selected structures to file
    distinct_structures_file = os.path.join(work_dir,  'distinct_structures.xyz')
    ase.io.write(distinct_structures_file, selected_atoms_list, format='extxyz')

    output = {
        'url': distinct_structures_file,
        'format': DataFormat.EXTXYZ,
        'attrs': attrs,
    }
    dump_json(output, os.path.join(work_dir, 'output.debug.json'))
    flush_stdio()  # flush joblib stdio buffer
    return output



from ai2_kit.core.script import BashTemplate, BashStep, BashScript
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.log import get_logger
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import list_split, dict_nested_get, dump_json, dump_text
from ai2_kit.core.pydantic import BaseModel
from ai2_kit.core.connector import get_ln_cmd

from typing import List, Literal, Optional, Mapping, Sequence, Any
from pydantic import validator, root_validator
from dataclasses import dataclass
from string import Template

import os
import itertools
import random
import ase.io


from .data import artifacts_to_ase_atoms_v2, DataFormat
from .iface import BaseCllContext, ICllExploreOutput
from .cp2k import dump_coord_n_cell
from .lammps import _get_dp_models_variables


class AnywareContextConfig(BaseModel):
    script_template: BashTemplate
    concurrency: int = 5


class AnywareConfig(BaseModel):

    system_files: List[str]
    """
    Artifact keys to the system
    """

    template_files: Mapping[str, str]
    """
    Templates files that will generate for each explore tasks,
    You can use $$VAR_NAME to reference the variables defined in product_vars and broadcast_vars.

    Besides, the following build-in variables are also available:
    - SYSTEM_FILE: the path of the system file
    - DP_MODELS: the path of the deep potential models, in the format of '1.pb 2.pb 3.pb 4.pb'

    You can use literal string to define the template file,
    or use !load_text to load the content from a file.

    For example, if you define a template file named 'cp2k.inp' with the following content:
        cp2k-warmup.inp: |
          &GLOBAL
          ...
          &END GLOBAL
        cp2k.inp: !load_text cp2k.inp

    """

    product_vars: Mapping[str, List[str]] = {}
    """
    Define template variables by Cartesian product
    The variable can be referenced in the template file with the following format:
    If there are too many variables, it will generate a large number of tasks,
    in this case, you can use broadcast_vars to reduce the number of tasks.

    $$VAR_NAME
    """

    broadcast_vars: Mapping[str, List[str]] = {}
    """
    Define template variables by broadcast (broadcast as in numpy).
    It's the same as product_vars, except that it will broadcast the variable to all other variables.
    """

    system_file_name: str
    """
    The name of the system file you want to generate,
    for example, 'system.xyz', 'coord_n_cell.inc', etc
    """

    system_file_format: str
    """
    The format of the system file you want to generate,
    for example, `lammps-data`, `cp2k-inc`,  etc

    For all supported data, you can refer to ase.io
    https://wiki.fysik.dtu.dk/ase/ase/io/io.html

    Custom formats:
    - cp2k-inc: coord & cell in the format of CP2K include file, can be used in CP2K input file via `@include coord_n_cell.inc`
    """

    submit_script: str
    """
    A bash script that will be executed in each task directory to submit the task.
    For example,

    mpirun cp2k.popt -i cp2k.inp &> cp2k.out
    """

    post_process_fn: Optional[str] = None
    """
    A python function that will be executed after the task is finished.
    You may use this function to post-process the results.

    The function must named as `post_process_fn` and accept a list of task directories as input.
    The below is an example of merging multiple file into one by keeping only the last line of each file.

    post_process_fn: |
        def post_process_fn(task_dirs):
            import glob
            for task_dir in task_dirs:
                files = glob.glob(os.path.join(task_dir, '*.out'))  # file to merge
                with open(os.path.join(task_dir, 'merged.out'), 'w') as fp:
                    for file in files:
                        with open(file, 'r') as f:
                            lines = f.readlines()
                            if len(lines) > 0:
                                fp.write(lines[-1])
    """

    delimiter: str = '$$'
    """
    delimiter for template
    """

    shuffle: bool = False
    """
    shuffle the combination of system_files, product_vars and broadcast_vars
    """

@dataclass
class AnywareInput:
    config: AnywareConfig
    new_system_files: List[Artifact]
    dp_models: Mapping[str, List[Artifact]]
    type_map: List[str]
    mass_map: List[float]


@dataclass
class AnywareContext(BaseCllContext):
    config: AnywareContextConfig


@dataclass
class AnywareOutput(ICllExploreOutput):
    output_dirs: List[Artifact]

    def get_model_devi_dataset(self) -> List[Artifact]:
        return self.output_dirs


async def anyware(input: AnywareInput, ctx: AnywareContext) -> AnywareOutput:
    executor = ctx.resource_manager.default_executor
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)

    if len(input.new_system_files) > 0:
        data_files = input.new_system_files
    else:
        data_files = ctx.resource_manager.resolve_artifacts(input.config.system_files)
    assert len(data_files) > 0, 'no data files found'

    task_artifacts = executor.run_python_fn(make_anyware_task_dirs)(
        work_dir=work_dir,
        data_files=data_files,
        dp_models={k: [m.url for m in v] for k, v in input.dp_models.items()},
        type_map=input.type_map,
        mass_map=input.mass_map,
        product_vars=input.config.product_vars,
        broadcast_vars=input.config.broadcast_vars,
        template_files=input.config.template_files,
        template_delimiter=input.config.delimiter,
        system_file_name=input.config.system_file_name,
        system_file_format=input.config.system_file_format,
        shuffle=input.config.shuffle,
    )
    steps = []
    for task_artifact in task_artifacts:
        steps.append(BashStep(
            cwd=task_artifact.url, cmd=input.config.submit_script, checkpoint='submit')
        )

    # # submit jobs by the number of concurrency
    jobs = []
    for i, steps_group in enumerate(list_split(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=work_dir)
        jobs.append(job)

    await gather_jobs(jobs, max_tries=2)

    if input.config.post_process_fn:
        executor.run_python_fn(run_post_process_fn)(
            post_process_fn=input.config.post_process_fn,
            task_dirs=[task.url for task in task_artifacts]
        )

    return AnywareOutput(output_dirs=task_artifacts)


def run_post_process_fn(post_process_fn: str, task_dirs: List[str]):
    _locals = {}
    exec(post_process_fn, None, _locals)
    _locals['post_process_fn'](task_dirs)


def make_anyware_task_dirs(work_dir: str,
                           data_files: List[Artifact],
                           dp_models: Mapping[str, List[str]],
                           type_map: List[str],
                           mass_map: List[float],
                           product_vars: Mapping[str, List[str]],
                           broadcast_vars: Mapping[str, List[str]],
                           template_files: Mapping[str, str],
                           template_delimiter: str,
                           system_file_name: str,
                           system_file_format: str,
                           shuffle: bool,
                           ):
    class _Template(Template):
        delimiter = template_delimiter

    # handle data files
    systems_dir = os.path.join(work_dir, 'systems')
    os.makedirs(systems_dir, exist_ok=True)
    atoms_list = []
    atoms_list = artifacts_to_ase_atoms_v2(data_files)

    system_artifacts = []
    for i, (artifact, atoms) in enumerate(atoms_list):
        ancestor = artifact.attrs['ancestor']
        data_file = os.path.join(systems_dir, f'{ancestor}-{i:06d}-{system_file_name}')
        if system_file_format == 'cp2k-inc':
            with open(data_file, 'w') as fp:
                dump_coord_n_cell(fp, atoms)
        elif system_file_format == 'lammps-data':
            ase.io.write(data_file, atoms, format=system_file_format, specorder=type_map)  # type: ignore
        else:
            ase.io.write(data_file, atoms, format=system_file_format)  # type: ignore
        system_artifacts.append(Artifact(url=data_file, attrs=artifact.attrs))
    if shuffle:
        random.shuffle(system_artifacts)

    # handle task dirs
    combination_fields: List[str] = ['SYSTEM_FILE']
    combination_values: List[List[Any]] = [system_artifacts]

    for k, v in product_vars.items():
        combination_fields.append(k)
        if shuffle:
            random.shuffle(v)
        combination_values.append(v)

    combinations = itertools.product(*combination_values)
    combinations = list(map(list, combinations))

    combination_fields.extend(broadcast_vars.keys())
    for i, combination in enumerate(combinations):
        for _vars in broadcast_vars.values():
            combination.append(_vars[i % len(_vars)])

    task_artifacts = []
    tasks_base_dir = os.path.join(work_dir, 'tasks')
    for i, combination in enumerate(combinations):
        task_dir = os.path.join(tasks_base_dir, f'{i:06d}')
        os.makedirs(task_dir, exist_ok=True)
        template_vars = dict(zip(combination_fields, combination))

        # link system_file to task_dir
        system_artifact: Artifact = template_vars.pop('SYSTEM_FILE')
        system_file = os.path.join(task_dir, system_file_name)
        # the reason of not using os.symlink is that it will raise an error if the link already exists
        os.system(get_ln_cmd(system_artifact.url, system_file))
        template_vars['SYSTEM_FILE'] = system_file_name
        # dp models variables
        dp_vars = _get_dp_models_variables(dp_models)
        # generate template files
        for k, v in template_files.items():
            template_file_path = os.path.join(task_dir, k)
            dump_text(_Template(v).substitute(**template_vars, **dp_vars), template_file_path, encoding='utf-8')
        task_artifacts.append(Artifact(url=task_dir,
                                       attrs={**system_artifact.attrs, 'source': system_artifact.url},
                                       format=DataFormat.ANYWARE_OUTPUT_DIR))
    return task_artifacts



from typing import TypeVar, Union, Callable, NamedTuple
from threading import Lock
import functools
import cloudpickle
import os
import inspect

from .log import get_logger
from .util import to_awaitable, ensure_dir

logger = get_logger(__name__)


class FnInfo(NamedTuple):
    fn_name: str
    args: tuple
    kwargs: dict
    call_site: str

KeyFn = Callable[[FnInfo], str]
EMPTY = object()

class CheckpointService:

    def __init__(self):
        self.checkpoint_dir = None
        self._lock = Lock()

    def set_checkpoint_dir(self, checkpoint_dir: str):
        self.checkpoint_dir = checkpoint_dir
        try:
            os.makedirs(checkpoint_dir, exist_ok=True)
        except FileExistsError:
            logger.error("ai2-kit use directory as checkpoint since 0.17.0")
            raise

    def apply_checkpoint(self, key_fn: Union[str, KeyFn], disable = False):
        call_site = inspect.getframeinfo(inspect.stack()[1][0])

        T = TypeVar('T', bound=Callable)
        def _checkpoint(fn: T) -> T:

            @functools.wraps(fn)
            def wrapper(*args, **kwargs):
                fn_info = FnInfo(
                    fn_name=fn.__name__,
                    args=args,
                    kwargs=kwargs,
                    call_site=f'{call_site.filename}:{call_site.lineno}',
                )

                key = key_fn if isinstance(key_fn, str) else key_fn(fn_info)

                if disable or self.checkpoint_dir is None:
                    return fn(*args, **kwargs)

                ret = self._get_checkpoint(key)
                if ret is not EMPTY:
                    logger.info(f"Hit checkpoint: {key}")
                    return ret

                ret = fn(*args, **kwargs)
                if inspect.isawaitable(ret):
                    async def _wrap_fn():
                        _ret = await ret
                        self._set_checkpoint(key, _ret, fn_info, True)
                        return _ret
                    return _wrap_fn()
                else:
                    self._set_checkpoint(key, ret, fn_info, False)
                    return ret

            return wrapper # type: ignore
        return _checkpoint

    def _get_checkpoint(self, key: str):
        assert self.checkpoint_dir is not None
        try:
            with self._lock:
                checkpoint_file = os.path.join(self.checkpoint_dir, key)
                if not os.path.exists(checkpoint_file):
                    return EMPTY
                with open(checkpoint_file, 'rb') as f:
                    value = cloudpickle.load(f)
                if value['is_awaitable']:
                    return to_awaitable(value['return'])
                else:
                    return value['return']
        except Exception as e:
            logger.exception(f"Fail to get checkpoint: {key}")
            return EMPTY


    def _set_checkpoint(self, key: str, value, info: FnInfo, is_awaitable: bool = False):
        assert self.checkpoint_dir is not None
        try:
            with self._lock:
                # args, kwargs may contain unpickable objects
                _checkpoint_data = {
                    'return': value,
                    'is_awaitable': is_awaitable,
                    'info': {
                        'fn_name': info.fn_name,
                        'call_site': info.call_site,
                    }
                }
                checkpoint_file = os.path.join(self.checkpoint_dir, key)
                ensure_dir(checkpoint_file)
                with open(checkpoint_file, 'wb') as f:
                    cloudpickle.dump(_checkpoint_data, f)

        except Exception as e:
            logger.exception('Fail to set checkpoint')


checkpoint_service = CheckpointService()


def set_checkpoint_dir(checkpoint_dir: str):
    checkpoint_service.set_checkpoint_dir(checkpoint_dir)


def apply_checkpoint(key_fn: Union[str, KeyFn], disable = False):
    """
    apply checkpoint for function.

    Note: This checkpoint implementation doesn't support multiprocess.
    To support multiple process we need to have a dedicated background process to read/write checkpoint,
    which will require message queue (e.g. nanomsg or nng) to implement it.

    Example:

    >>> set_checkpoint_file('/tmp/test.ckpt')
    >>> task_fn = lambda a, b: a + b
    >>> checkpoint('task_1+2')(task_fn)(1, 2)
    """

    return checkpoint_service.apply_checkpoint(key_fn, disable)



from typing import Union, Iterable, List
from collections import namedtuple
from scipy.stats import gaussian_kde
import numpy as np
import pandas as pd

import dpdata

default_kB = 0.0083144621  # Boltzmann constant in kJ/mol/K
default_ev_to_kjmol = 96.4853365  # conversion factor from eV to kJ/mol


_FesResult = namedtuple('FesResult', ['fes', 'grid', 'extend'])

def compute_fes(cvs: np.ndarray, bias: np.ndarray, temp: float, grid=None,
                w=1, kB=default_kB, grid_size=100j):
    """
    Compute free energy surface from biased sampling data using Gaussian KDE
    Support 1D and 2D collective variables

    :param cv: collective variable
    :param bias: bias potential
    :param temp: temperature
    :param grid: grid points, if None, will use grid_size to generate automatically
    :param w: weights
    :param kB: Boltzmann constant
    :param grid_size: grid size in np.mgrid style

    :return: free energy
    """
    if cvs.ndim == 2:
        if cvs.shape[0] == 1:
            cvs = cvs[0]  # flatten
        elif cvs.shape[0] != 2:
            raise ValueError(f'Invalid cvs shape {cvs.shape}, only support 1D or 2D cvs')
    elif cvs.ndim != 1:
        raise ValueError(f'Invalid cvs shape {cvs.shape}, only support 1D or 2D cvs')

    kBT = kB * temp
    beta = 1 / kBT
    weights = w * np.exp(beta * bias)
    kde = gaussian_kde(cvs, weights=weights)

    extend = None
    if cvs.ndim == 1:  # handle 1d
        if grid is None:
            x_min, x_max = cvs.min(), cvs.max()
            grid = np.mgrid[x_min:x_max:grid_size]
            extend = [x_min, x_max]
        pdf = kde.evaluate(grid)
    else:  # handle 2d
        if grid is None:
            x_min, x_max = cvs[0].min(), cvs[0].max()
            y_min, y_max = cvs[1].min(), cvs[1].max()
            grid = np.mgrid[x_min:x_max:grid_size, y_min:y_max:grid_size]
            extend = [x_min, x_max, y_min, y_max]
        X, Y = grid
        pos = np.vstack([X.ravel(), Y.ravel()])
        # TODO: generated by copilot, I am not sure if it is correct
        pdf = kde.evaluate(pos).T.reshape(X.shape)
    fes = -kBT * np.log(pdf)
    # return fes with grid and extend for plotting
    return _FesResult(fes=fes, grid=grid, extend=extend)




def compute_kde_weight(baseline_energy: np.ndarray, target_energy: np.ndarray,
                       temp: float, kB=default_kB,
                       ev_to_kjmol=default_ev_to_kjmol,):
    """
    Compute KDE weights for reweighting

    :param baseline_energy: baseline energy in eV, it's the output of DeepMD evaluation
    :param target_energy: target energy in eV, it's the output of DeepMD evaluation
    """
    kBT = kB * temp
    beta = 1 / kBT

    # use relative energy and convert from eV to kJ/mol
    be = (baseline_energy- np.min(baseline_energy)) * ev_to_kjmol
    te = (target_energy- np.min(target_energy)) * ev_to_kjmol
    return np.exp( -beta * (te - be))



class ReweightTool:

    @classmethod
    def from_dpdata_system(cls, dp_sys):
        ...


    def __init__(self):
        ...


from fire import Fire


class Group:
    def __init__(self, items: dict, doc: str = '') -> None:
        self.__doc__ = doc
        self.__dict__.update(items)


class AlgorithmGroup:
    """
    Algorithms for specific domains.
    """

    def proton_transfer(self):
        """
        Proton transfer analysis toolkit.
        """
        from ai2_kit.algorithm import proton_transfer
        return proton_transfer.cli_entry

    def aosa(self):
        """
        Amorphous oxides structure analysis toolkit.
        """
        from ai2_kit.algorithm import aos_analysis
        return aos_analysis.cli_entry


class WorkflowGroup:
    """
    Workflows for specific domains.
    """

    @property
    def cll_mlp_training(self):
        from ai2_kit.workflow.cll_mlp import run_workflow
        return run_workflow

    @property
    def fep_mlp_training(self):
        from ai2_kit.workflow.fep_mlp import run_workflow
        return run_workflow

class ToolGroup:
    """
    Tools for specific domains.
    """

    @property
    def ase(self):
        from ai2_kit.tool.ase import AseTool
        return AseTool

    @property
    def dpdata(self):
        from ai2_kit.tool.dpdata import DpdataTool
        return DpdataTool

    @property
    def yaml(self):
        from ai2_kit.tool.yaml import Yaml
        return Yaml

    @property
    def batch(self):
        from ai2_kit.tool.batch import BatchTool
        return BatchTool

    @property
    def hpc(self):
        from ai2_kit.tool.hpc import cmd_entry
        return cmd_entry

    @property
    def misc(self):
        from ai2_kit.tool.misc import cmd_entry
        return cmd_entry

class FeatureGroup:
    """
    Featuring tools for specific domains.
    """

    @property
    def catalysis(self):
        """
        Catalyst specific tools.
        """
        from ai2_kit.feat.catalysis import CmdEntries
        return CmdEntries

    @property
    def cat(self):
        """
        Shortcut for catalyst.
        """
        return self.catalysis

    @property
    def spectr(self):
        """
        Spectrum specific tools.
        """
        from ai2_kit.feat.spectrum import CmdEntries
        return CmdEntries()

ai2_kit = Group({
    'workflow': WorkflowGroup(),
    'algorithm': AlgorithmGroup(),
    'tool': ToolGroup(),
    'feat': FeatureGroup(),

}, doc="Welcome to use ai2-kit!")


def main():
    Fire(ai2_kit)


if __name__ == '__main__':
    main()



from .pydantic import BaseModel
from typing import Optional, Mapping, TypedDict
import os
import copy


class ArtifactDict(TypedDict):
    """
    A dict representation of Artifact.
    Use this when you need to run a remote function call as pydantic model is not pickleable.

    referrer is not included in this dict as it is not pickleable.
    """
    url: str
    attrs: dict
    executor: Optional[str]
    format: Optional[str]
    includes: Optional[str]
    key: Optional[str]


class Artifact(BaseModel):

    key: Optional[str] = None
    executor: Optional[str] = None
    url: str
    format: Optional[str] = None
    includes: Optional[str] = None
    attrs: dict = dict()

    @classmethod
    def of(cls,
           url: str,
           key: Optional[str] = None,
           executor: Optional[str] = None,
           includes: Optional[str] = None,
           attrs: Optional[dict] = None,
           format: Optional[str] = None,):
        """
        Create an Artifact instance. Use this instead of __init__ to avoid type error of pydantic
        This method should be deprecated.
        """
        return cls(url=url,
                   executor=executor,
                   format=format,
                   includes=includes,
                   attrs=dict() if attrs is None else copy.deepcopy(attrs),  # deepcopy to reduce chance of mistake
                   key=key)

    def to_dict(self) -> ArtifactDict:
        """Convert to a dict representation.
        Use this when you need to run a remote function call as pydantic model is not pickleable."""
        return self.dict() # type: ignore

    def join(self, *paths, **kwargs):
        url = os.path.join(self.url, *paths)
        return Artifact.of(url=url, executor=self.executor, **kwargs)


ArtifactMap = Mapping[str, Artifact]
